import asyncio
import logging
from enum import Enum
from typing import Optional, Awaitable, Callable

from .session import SessionLike
from .data_structure import Request, Response, SessionOptions
from .dslv2 import parse as parse_v2
from .dslv1 import parse as parse_v1
from .excs import FuoSyntaxError

logger = logging.getLogger(__name__)


class ProtocolType(Enum):
    rpc = 'rpc'
    pubsub = 'pubsub'


class DeadSubscriber(Exception):
    pass


# Internal exceptions.
class RequestError(Exception):
    pass


class _EOF(Exception):
    pass


def encode(s):
    return bytes(s, 'utf-8')


def decode(b):
    return b.decode('utf-8')


async def read_request(reader: asyncio.StreamReader, parse_func) -> Optional[Request]:
    """读取一个请求

    读取成功时，返回 Request 对象，如果读取失败（比如客户端关闭连接），
    则返回 None。其它异常会抛出 RequestError。

    :type reader: asyncio.StreamReader
    """
    try:
        line_bytes = await reader.readline()
    except ValueError as e:
        raise RequestError('request size should be less than 64KiB') from e
    if not line_bytes:  # EOF
        raise _EOF
    line_text = line_bytes.decode('utf-8').strip()
    if not line_text:
        return None
    try:
        req = parse_func(line_text)
    except FuoSyntaxError as e:
        raise RequestError(e.human_readabe_msg) from e
    else:
        if not req.has_heredoc:
            return req
        word_bytes = bytes(req.heredoc_word, 'utf-8')
        buf = bytearray()
        while 1:
            line_bytes = await reader.readline()
            if line_bytes[-2:] == b'\r\n':
                stripped_line_bytes = line_bytes[:-2]
            else:
                stripped_line_bytes = line_bytes[:-1]
            if stripped_line_bytes == word_bytes:
                break
            buf.extend(line_bytes)
            if len(buf) >= 2 ** 16:
                raise RequestError('heredoc body should be less than 64KiB')
        req.set_heredoc_body(bytes(buf).decode('utf-8'))
    return req


class FuoServerProtocol(asyncio.streams.FlowControlMixin):
    """asyncio-style fuo server protocol (ClientHandler)

    Implementation references:
    - asyncio.streams.StreamReaderProtocol
    - aiohttp.web_protocol.RequestHandler

    TODO:
    - add request timeout: close connection if no action happens
    - add graceful shutdown: close connection before exit
    """
    def __init__(
            self,
            handle_req: Callable[[Request, SessionLike], Awaitable[Response]],
            loop):
        super().__init__(loop)
        self._handle_req = handle_req
        self._loop = loop

        self.options: SessionOptions = SessionOptions()

        # StreamReader provides some convinient file-object-like methods
        # like readline, which is really useful for our implementation.
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None

        self._peername = None

    #
    # writer and reader property can only be used after connection is made.
    #
    @property
    def writer(self):
        assert self._writer is not None
        return self._writer

    @property
    def reader(self):
        assert self._reader is not None
        return self._reader

    async def read_request(self):
        if self.options.rpc_version == '2.0':
            parse_func = parse_v2
        else:
            parse_func = parse_v1
        return await read_request(self.reader, parse_func)

    async def write_welcome(self):
        version = self.options.rpc_version
        self.writer.write(encode(f'OK rpc {version}\r\n'))

    async def write_response(self, resp):
        # TODO: 区分客户端和服务端错误（比如客户端错误后面加 ! 标记）
        msg_bytes = bytes(resp.text, 'utf-8')
        response_line = f'ACK {resp.code} {len(msg_bytes)}\r\n'
        self.writer.write(bytes(response_line, 'utf-8'))
        self.writer.write(msg_bytes)
        self.writer.write(b'\r\n')
        await self.writer.drain()

    async def start(self):
        """connection handler"""
        # we should call drain after each write to do flow control,
        # though it is not so important in this case.
        try:
            await self.write_welcome()
            await self.writer.drain()
            # HELP: static checker say there is no attribute "_connection_lost".
            while not self._connection_lost:  # type: ignore
                try:
                    req = await self.read_request()
                except RequestError as e:
                    msg = 'bad reqeust!\r\n' + str(e)
                    bad_request_resp = Response(ok=False, text=msg)
                    await self.write_response(bad_request_resp)
                except _EOF:
                    # Client closed the connection.
                    break
                else:
                    if req is None:
                        # Ignore the empty request.
                        continue

                    # 通常来说，客户端如果想断开连接，只需要自己主动关闭连接即可，
                    # 但如果客户端不方便主动断开，可以发送 quit 命令，
                    # 让服务端来主动关闭连接。
                    #
                    # 客户端不方便主动断开的例子：当用户使用 NetCat 发送请求时，
                    # 用户想在读取完一个 fuo 响应后断开。
                    if req.cmd == 'quit':
                        # FIXME: 理论上最好能等待 close 结束
                        self.writer.close()
                    else:
                        try:
                            resp = await self._handle_req(req, self)
                        except Exception as e:  # pylint: disable=broad-except
                            msg = f'server error!\r\n{repr(e)}'
                            resp = Response(ok=False, text=msg, req=req)
                        await self.write_response(resp)
        except ConnectionResetError:
            # client close the connection
            pass

    def connection_made(self, transport):
        self._peername = transport.get_extra_info('peername')
        logger.debug('%s connceted to fuo daemon.', self._peername)
        self._reader = asyncio.StreamReader(loop=self._loop)
        self._writer = asyncio.StreamWriter(
            transport, self, self._reader, self._loop)
        # Unlike aiohttp RequestHandler, we will never cancel the handler task,
        # our task should die when it is supposed to. For instance, when the
        # client close the connection, connection lost with ConnectionResetErrror,
        # the handler task should catch the exc and exit.
        self._loop.create_task(self.start())

    def connection_lost(self, exc):
        """called when our transport is closed"""
        if self._reader is not None:
            if exc is None:
                self._reader.feed_eof()
            else:
                self._reader.set_exception(exc)
        super().connection_lost(exc)
        # HELP: if you dive into aiohttp RequestHandler or StreamReaderProtocol,
        # you can see that they set reader, writer...almost
        # every thing to None when connection lost. I just dont know why.
        # I have done some experiment, the reader and writer can be
        # gc-collected even if we do not set it to None manually.
        # they can be deleted after protocol(self) is deleted.
        self._reader = None
        self._writer = None
        logger.debug('%s disconnceted from fuo daemon.', self._peername)

    def data_received(self, data):
        self.reader.feed_data(data)

    def eof_received(self):
        """client has written eof

        Explaination: this means the client will send no more data,
        client has close the socket or shutdown with SHUT_WR/SHUT_RDWR flag.
        If we use tcpdump to do packet capture, we can see a FIN packet.

        For fuo protocol, if client shutdown the write pipe, we believe that
        the client is closing the socket, we will just close the socket,
        send no data any more, even if the last response may not complete.
        """
        return False

    def write_topic_msg(self, topic, msg):
        """
        TODO: Create a enum for version.
        """
        if self._connection_lost or self.writer.is_closing():  # type: ignore
            raise DeadSubscriber

        body = encode(msg)
        try:
            if self.options.pubsub_version == '2.0':
                response_line = f'MSG {topic} {len(body)}\r\n'
                self.writer.write(encode(response_line))
                self.writer.write(body)
                self.writer.write(b'\r\n')
            else:
                self.writer.write(body)
        except BrokenPipeError:
            # What's happening?
            raise DeadSubscriber from None


class PubsubProtocol(FuoServerProtocol):

    async def write_welcome(self):
        version = self.options.pubsub_version
        self.writer.write(encode(f'OK pubsub {version}\r\n'))



from PyQt5.QtCore import QTimer, QSize
from PyQt5.QtGui import QColor, QPalette
from PyQt5.QtWidgets import QVBoxLayout, QPushButton, QWidget, \
    QHBoxLayout, QOpenGLWidget, QLabel

from feeluown.player import State
from feeluown.gui.widgets import PlayPauseButton
from feeluown.gui.widgets.progress_slider import ProgressSlider
from feeluown.gui.widgets.size_grip import SizeGrip
from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.widgets.labels import ProgressLabel, DurationLabel


class Button(TextButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        font = self.font()
        font.setPixelSize(10)
        self.setFont(font)
        self.setFixedHeight(16)

    def sizeHint(self):
        fm = self.fontMetrics()
        rect = fm.boundingRect(self.text())
        return QSize(rect.width() + 6, rect.height())


class VideoPlayerCtlBar(QWidget):

    def __init__(self, app, parent=None):
        super().__init__(parent=parent)

        self._app = app

        # Create widgets.
        self._toggle_btn = PlayPauseButton(draw_circle=False)
        self._progress_slider = ProgressSlider(app)
        self._progress_label = ProgressLabel(app, self)
        self._duration_label = DurationLabel(app, self)
        self._size_grip = SizeGrip(parent=self)
        self._adhoc_btn_layout = QHBoxLayout()
        self._layout = QVBoxLayout(self)
        self._bottom_layout = QHBoxLayout()

        # Do initialization for widgets.
        self._toggle_btn.setCheckable(True)
        self._setup_ui()

        self._app.player.state_changed.connect(
            lambda x: self._toggle_btn.setChecked(x == State.playing),
            weak=False,
            aioqueue=True)
        self._toggle_btn.clicked.connect(self._app.player.toggle)

    def add_adhoc_btn(self, text: str) -> QPushButton:
        """
        .. versionadded: 4.0.1
        """
        btn = Button(text)
        self._adhoc_btn_layout.addWidget(btn)
        btn.setPalette(self.palette())
        return btn

    def clear_adhoc_btns(self):
        """
        .. versionadded: 4.0.1
        """
        while self._adhoc_btn_layout.count():
            item = self._adhoc_btn_layout.itemAt(0)
            self._adhoc_btn_layout.removeItem(item)
            if item.widget():
                item.widget().deleteLater()

    def _setup_ui(self):
        self.setAutoFillBackground(True)

        # Setup layout.
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self._progress_slider)
        self._layout.addLayout(self._bottom_layout)

        # Setup bottom layout.
        self._layout.addStretch(0)
        self._bottom_layout.setSpacing(2)
        self._adhoc_btn_layout.setSpacing(4)
        self._bottom_layout.addWidget(self._toggle_btn)
        self._bottom_layout.addSpacing(6)
        self._bottom_layout.addWidget(self._progress_label)
        self._bottom_layout.addWidget(QLabel('/', self))
        self._bottom_layout.addWidget(self._duration_label)
        self._bottom_layout.addStretch(1)
        self._bottom_layout.addLayout(self._adhoc_btn_layout)
        self._bottom_layout.addSpacing(6)
        self._bottom_layout.addWidget(self._size_grip)

        # Setup widgets size.
        self._size_grip.setFixedSize(20, 20)

        # Customize the palette.
        palette = self.palette()
        palette.setColor(QPalette.Text, QColor('white'))
        palette.setColor(QPalette.ButtonText, QColor('white'))
        palette.setColor(QPalette.WindowText, QColor('white'))
        bg_color = QColor('black')
        bg_color.setAlpha(180)  # Make it semi-transparent.
        palette.setColor(QPalette.Window, bg_color)
        self.setPalette(palette)
        # It seems children don't inherit the palette, so we set palette for
        # children explicitly.
        for child in self.findChildren(QWidget):
            child.setPalette(palette)


class VideoOpenGLWidget(QOpenGLWidget):
    """Base class for widget which provides OpenGL based video rendering

    This widget has a overlay which currently contains a player control bar.
    """
    def __init__(self, app, parent=None):
        super().__init__(parent=parent)

        # Define public attributes.
        self.overlay_auto_visible = True

        # Define protected attributes.
        self._gl_painters = set({})
        self._overlay_visible_timer = QTimer(self)
        self._overlay_visible_duration = 2000  # 2s

        # Define ui related objects and attributes.
        self.ctl_bar = VideoPlayerCtlBar(app=app, parent=self)
        self._layout = QVBoxLayout(self)

        # Do initialization.
        self.setMouseTracking(True)
        self._overlay_visible_timer.timeout.connect(self.hide_overlay)
        self._setup_ui()

    def _setup_ui(self):
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addStretch(1)
        self._layout.addWidget(self.ctl_bar)
        self.hide_overlay()

    def add_gl_painter(self, painter):
        """Add QPainter based gl painter.

        In paintGL, we invoke these painters one by one so that they can
        paint their own contents.

        :param painter: painter must impl `paint(opengl_widget)` method.

        .. versionadded: 3.7.13
        """
        # Simple validation
        assert hasattr(painter, 'paint'), \
            "painter must impl `paint` method."
        self._gl_painters.add(painter)

    def hide_overlay(self):
        """hide the overlay widget"""
        self.ctl_bar.hide()

    def show_overlay(self):
        """show the overlay widget"""
        self.ctl_bar.show()

    def mouseMoveEvent(self, e):
        super().mouseMoveEvent(e)
        if self.overlay_auto_visible is True:
            self.show_overlay()
            self._overlay_visible_timer.start(self._overlay_visible_duration)

    def enterEvent(self, e):
        super().enterEvent(e)
        if self.overlay_auto_visible is True:
            self.show_overlay()
            self._overlay_visible_timer.start(self._overlay_visible_duration)

    def leaveEvent(self, e):
        super().leaveEvent(e)
        self._overlay_visible_timer.start(self._overlay_visible_duration)


if __name__ == '__main__':
    from feeluown.gui.debug import simple_layout, mock_app

    with simple_layout(theme='dark') as layout, mock_app() as app:
        ctl_bar = VideoPlayerCtlBar(app)
        ctl_bar.add_adhoc_btn('全屏')
        ctl_bar.add_adhoc_btn('最小化')
        layout.addWidget(ctl_bar)



from typing import TYPE_CHECKING

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QFrame, QLabel, QVBoxLayout, QHBoxLayout, QFormLayout,
    QDialog, QLineEdit, QDialogButtonBox, QMessageBox, QWidget, QSizePolicy
)

from feeluown.excs import ProviderIOError, NoUserLoggedIn
from feeluown.library import (
    SupportsPlaylistDelete, SupportsPlaylistCreateByName, SupportsCurrentUser,
)
from feeluown.utils import aio
from feeluown.gui.provider_ui import UISupportsNavBtns, UISupportsCreatePlaylist
from feeluown.gui.components import Avatar
from feeluown.gui.widgets import (
    DiscoveryButton, StarButton, PlusButton, TriagleButton,
    EmojiButton,
)
from feeluown.gui.widgets.playlists import PlaylistsView
from feeluown.gui.widgets.my_music import MyMusicView

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class ListViewContainer(QFrame):

    def __init__(self, label, view, parent=None):
        super().__init__(parent)

        self._btn_length = 14
        self._label = label
        self._view = view
        self._toggle_btn = TriagleButton(length=self._btn_length, padding=0.2)
        self.create_btn = PlusButton(length=self._btn_length)
        # Show this button when needed.
        self.create_btn.hide()

        self._toggle_btn.clicked.connect(self.toggle_view)
        self.setup_ui()

    def setup_ui(self):
        self._label.setFixedHeight(25)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        self._t_h_layout = QHBoxLayout()
        self._b_h_layout = QHBoxLayout()
        self._t_h_layout.addWidget(self._label)
        self._t_h_layout.addStretch(0)
        self._t_h_layout.addWidget(self.create_btn)
        self._t_h_layout.addSpacing(self._btn_length // 2)
        self._t_h_layout.addWidget(self._toggle_btn)
        self._b_h_layout.addWidget(self._view)

        self._layout.addLayout(self._t_h_layout)
        self._layout.addLayout(self._b_h_layout)
        # XXX: 本意是让 ListViewContainer 下方不要出现多余的空间
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)

    def toggle_view(self):
        if self._view.isVisible():
            self._toggle_btn.set_direction('down')
            self._view.hide()
        else:
            self._toggle_btn.set_direction('up')
            self._view.show()


LVC = ListViewContainer


class ProviderBar(QWidget):
    """
    A panel shows provider-specific contents.
    """
    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent)
        self._app = app

        self.discovery_btn = DiscoveryButton(height=30, padding=0.2, parent=self)
        self.fav_btn = StarButton('我的收藏', height=30, parent=self)
        self.fold_top_btn = TriagleButton(length=14, padding=0.2)
        self.fold_top_btn.setCheckable(True)

        self.playlists_header = QLabel('歌单列表', self)
        self.my_music_header = QLabel('我的音乐', self)

        self._layout = QVBoxLayout(self)
        # Layout to let provider add it's own buttons.
        self._btn_layout = QVBoxLayout()
        self.playlists_view = PlaylistsView(self)
        self.my_music_view = MyMusicView(self)
        self.playlists_view.setModel(self._app.pl_uimgr.model)
        self.my_music_view.setModel(self._app.mymusic_uimgr.model)

        self.playlists_con = LVC(self.playlists_header, self.playlists_view)
        self.my_music_con = LVC(self.my_music_header, self.my_music_view)

        self.playlists_view.show_playlist.connect(
            lambda pl: self._app.browser.goto(model=pl))
        self.playlists_view.remove_playlist.connect(self._remove_playlist)
        self.playlists_con.create_btn.clicked.connect(self._create_playlist)
        self._app.current_pvd_ui_mgr.changed.connect(
            self.on_current_pvd_ui_changed)
        self.discovery_btn.clicked.connect(
            lambda: self._app.browser.goto(page='/rec'))
        self.fav_btn.clicked.connect(
            lambda: self._app.browser.goto(page='/my_fav'))

        self.setup_ui()

    def setup_ui(self):
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

        self.playlists_view.setFrameShape(QFrame.NoFrame)
        self.my_music_view.setFrameShape(QFrame.NoFrame)

        self._avatar_layout = QHBoxLayout()
        self._avatar_layout.addWidget(Avatar(self._app, height=48))
        self._avatar_layout.addWidget(self.fold_top_btn)

        self._layout.addLayout(self._avatar_layout)
        self._layout.addWidget(self.discovery_btn)
        self._layout.addWidget(self.fav_btn)
        self._layout.addLayout(self._btn_layout)
        self._layout.addWidget(self.my_music_con)
        self._layout.addWidget(self.playlists_con)

        # 让各个音乐库来决定是否显示这些组件
        self.playlists_con.hide()
        self.my_music_con.hide()
        self.discovery_btn.setDisabled(True)
        self.fav_btn.setDisabled(True)
        self.discovery_btn.setToolTip('当前资源提供方未知')
        self.fold_top_btn.setToolTip('折叠/打开“主页和本地收藏集”功能')

    def on_current_pvd_ui_changed(self, pvd_ui, _):
        self._clear_btns()
        if pvd_ui:
            self.discovery_btn.setEnabled(True)
            self.fav_btn.setEnabled(True)
            self.discovery_btn.setToolTip(f'点击进入 {pvd_ui.provider.name} 推荐页')
            if isinstance(pvd_ui, UISupportsNavBtns):
                for btn in pvd_ui.list_nav_btns():
                    qt_btn = EmojiButton(btn.icon, btn.text, height=30, parent=self)
                    qt_btn.clicked.connect(btn.cb)
                    self._add_btn(qt_btn)

            if isinstance(pvd_ui, UISupportsCreatePlaylist):
                self.playlists_con.create_btn.show()
                self.playlists_con.create_btn.clicked.connect(pvd_ui.create_playlist)
            else:
                self.playlists_con.create_btn.hide()
        else:
            self.discovery_btn.setEnabled(False)
            self.fav_btn.setEnabled(False)

    def _add_btn(self, btn):
        self._btn_layout.addWidget(btn)

    def _clear_btns(self):
        for _ in range(self._btn_layout.count()):
            item = self._btn_layout.takeAt(0)
            item.widget().deleteLater()
            del item

    def _create_playlist(self):
        provider_ui = self._app.current_pvd_ui_mgr.get()
        if provider_ui is None:
            self._app.show_msg('当前的资源提供方未注册其 UI')
            return
        provider = provider_ui.provider
        if not isinstance(provider, SupportsPlaylistCreateByName) \
           or not isinstance(provider, SupportsCurrentUser) \
           or not provider.has_current_user():
            self._app.show_msg('当前的资源提供方不支持创建歌单')
            return

        dialog = QDialog(self)
        # Set WA_DeleteOnClose so that the dialog can be deleted (from self.children).
        dialog.setAttribute(Qt.WA_DeleteOnClose)
        layout = QFormLayout(dialog)
        title_edit = QLineEdit(dialog)
        layout.addRow('歌单名', title_edit)
        button_box = QDialogButtonBox(QDialogButtonBox.Cancel | QDialogButtonBox.Yes)
        layout.addRow('', button_box)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)

        def create_playlist_and_reload():
            title = title_edit.text()

            async def do():
                try:
                    playlist = await aio.run_fn(provider.playlist_create_by_name, title)
                except (ProviderIOError, NoUserLoggedIn) as e:
                    QMessageBox.warning(self._app, '错误', f"创建歌单 '{title}' 失败: {e}")
                else:
                    # Add playlist to pl_uimgr is a workaround, which may cause bug.
                    # For example, the order of the newly created playlist should be
                    # in the top for some providers.
                    # TODO: re-fetch user's playlists and fill the UI.
                    self._app.pl_uimgr.add(playlist, is_fav=False)
                    self._app.show_msg(f"创建歌单 '{title}' 成功")

            aio.run_afn(do)

        dialog.accepted.connect(create_playlist_and_reload)
        dialog.open()

    def _remove_playlist(self, playlist):

        async def do():
            provider = self._app.library.get(playlist.source)
            if isinstance(provider, SupportsPlaylistDelete):
                ok = await aio.run_fn(provider.playlist_delete, playlist.identifier)
                self._app.show_msg(f"删除歌单 {playlist.name} {'成功' if ok else '失败'}")
                if ok is True:
                    self._app.pl_uimgr.model.remove(playlist)
            else:
                self._app.show_msg(f'资源提供方({playlist.source})不支持删除歌单')

        box = QMessageBox(QMessageBox.Warning, '提示', f"确认删除歌单 '{playlist.name}' 吗？",
                          QMessageBox.Yes | QMessageBox.No, self)
        box.accepted.connect(lambda: aio.run_afn(do))
        box.open()



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QDialog, QWidget, QCheckBox, \
    QVBoxLayout, QHBoxLayout

from feeluown.gui.widgets.magicbox import KeySourceIn
from feeluown.gui.widgets.header import MidHeader
from feeluown.gui.components import LyricButton, WatchButton


class _ProviderCheckBox(QCheckBox):
    def set_identifier(self, identifier):
        self.identifier = identifier  # pylint: disable=W0201


class SearchProvidersFilter(QWidget):
    checked_btn_changed = pyqtSignal(list)

    def __init__(self, providers):
        super().__init__()
        self.providers = providers

        self._btns = []
        self._layout = QHBoxLayout(self)

        for provider in self.providers:
            btn = _ProviderCheckBox(provider.name, self)
            btn.set_identifier(provider.identifier)
            btn.clicked.connect(self.on_btn_clicked)
            self._layout.addWidget(btn)
            self._btns.append(btn)

        # HELP: we add spacing between checkboxes because they
        # will overlay each other on macOS by default. Why?
        self._layout.setSpacing(10)
        self._layout.addStretch(0)

    def get_checked_providers(self):
        identifiers = []
        for btn in self._btns:
            if btn.isChecked():
                identifiers.append(btn.identifier)
        return identifiers

    def set_checked_providers(self, providers):
        for provider in providers:
            for btn in self._btns:
                if provider == btn.identifier:
                    btn.setChecked(True)
                    break

    def on_btn_clicked(self, _):
        self.checked_btn_changed.emit(self.get_checked_providers())


class PlayerSettings(QWidget):
    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._app = app
        self.lyric_btn = LyricButton(app, height=16)
        self.watch_btn = WatchButton(app, height=16)
        self._layout = QHBoxLayout(self)
        self._layout.addWidget(self.lyric_btn)
        self._layout.addWidget(self.watch_btn)
        self._layout.addStretch(0)


class SettingsDialog(QDialog):
    def __init__(self, app, parent=None):
        super().__init__(parent=parent)
        self._app = app

        self.setWindowTitle('应用配置')
        self.render()

    def render(self):
        source_in_str = self._app.browser.local_storage.get(KeySourceIn)
        if source_in_str is not None:
            source_in = source_in_str.split(',')
        else:
            source_in = [p.identifier for p in self._app.library.list()]
        toolbar = SearchProvidersFilter(self._app.library.list())
        toolbar.set_checked_providers(source_in)
        toolbar.checked_btn_changed.connect(self.update_source_in)

        self._layout = QVBoxLayout(self)
        self._layout.addWidget(MidHeader('搜索来源'))
        self._layout.addWidget(toolbar)
        self._layout.addWidget(MidHeader('播放器'))
        self._layout.addWidget(PlayerSettings(self._app))
        self._layout.addStretch(0)

    def update_source_in(self, source_in):
        self._app.browser.local_storage[KeySourceIn] = ','.join(source_in)



import logging
from typing import Optional, TYPE_CHECKING

from PyQt5.QtCore import Qt

from feeluown.excs import ProviderIOError
from feeluown.utils.aio import run_fn, run_afn
from feeluown.player import SongRadio
from feeluown.library import SongModel, VideoModel, SearchType
from feeluown.gui.widgets.magicbox import KeySourceIn

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


MV_BTN_TEXT = '播放 MV'


class SongMenuInitializer:

    def __init__(self, app: 'GuiApp', song):
        """
        :type app: feeluown.app.App
        """
        self._app = app
        self._song = song
        self._fetching_artists = False
        self._fetching_mv = False

    def apply(self, menu):

        app = self._app
        song = self._song

        def enter_song_radio(song):
            radio = SongRadio.create(app, song)
            app.fm.activate(radio.fetch_songs_func, reset=False)
            if app.playlist.current_song != song:
                app.playlist.clear()
                self._app.playlist.next()
                self._app.player.resume()
            else:
                for song_ in app.playlist.list().copy():
                    if song_ is not app.playlist.current_song:
                        app.playlist.remove(song_)
                self._app.player.resume()

        def goto_song_explore(song):
            app.browser.goto(model=song, path='/explore')

        async def goto_song_album(song):
            usong: SongModel = await run_fn(self._app.library.song_upgrade, song)
            if usong.album is not None:
                self._app.browser.goto(model=usong.album)
            else:
                self._app.show_msg('该歌曲没有专辑信息')

        menu.hovered.connect(self.on_action_hovered)
        menu.addAction('搜索相似资源').triggered.connect(
            lambda: self.show_similar_resource(song))
        artist_menu = menu.addMenu('查看歌手')
        artist_menu.menuAction().setData({'artists': None, 'song': song})
        mv_menu = menu.addMenu(MV_BTN_TEXT)
        mv_menu.menuAction().setData({'mvs': None, 'song': song})

        menu.addAction('查看专辑').triggered.connect(
            lambda: run_afn(goto_song_album, song))
        menu.addAction('歌曲电台').triggered.connect(
            lambda: enter_song_radio(song))
        menu.addAction('歌曲详情').triggered.connect(
            lambda: goto_song_explore(song))

    def show_similar_resource(self, song):
        from feeluown.gui.components.search import SearchResultView

        # TODO: add a close button, and rename it to ClosableSearchResultView
        class SearchResultViewWithEsc(SearchResultView):
            def keyPressEvent(self, event):
                if event.key() == Qt.Key_Escape:
                    self.close()
                else:
                    super().keyPressEvent(event)

        q = f'{song.title} {song.artists_name}'
        view = SearchResultViewWithEsc(self._app, transparent_bg=False, parent=self._app)

        view.setStyleSheet('''
            SearchResultView {
                border: 1px solid gray;
                border-radius: 2px;
            }
        ''')

        source_in = self._app.browser.local_storage.get(KeySourceIn, None)
        run_afn(view.search_and_render, q, SearchType.so, source_in)

        width = self._app.width() - self._app.ui.sidebar.width()
        height = self._app.height() * 3 // 5
        x = self._app.ui.sidebar.width()
        y = self._app.height() - height - self._app.ui.player_bar.height()

        # Do not use Qt.Popup, because it behaves differently on different
        # platforms. For example, on macOS, it will pop up a window with a
        # shadow, while on Linux(KDE) it won't.
        view.setGeometry(x, y, width, height)
        view.show()
        view.raise_()

    def on_action_hovered(self, action):
        """
        Fetch song.artists when artists_action is hovered. If it is
        already fetched, ignore.
        """
        data = action.data()
        if data is None:  # submenu action
            return
        if 'artists' in data:
            self._hover_artists(action, data)
        elif 'mvs' in data:
            self._hover_mv(action, data)

    def _hover_mv(self, action, data):

        def mv_fetched_cb(future):
            self._fetching_mv = False
            try:
                mv: Optional[VideoModel] = future.result()
            except ProviderIOError as e:
                logger.error(f"fetch song mv failed {e}")
                mv = None
            if mv is not None:
                try:
                    mv_action = action.menu().addAction(mv.title)
                    mv_action.triggered.connect(
                        lambda: self._app.playlist.play_model(mv))
                except RuntimeError:
                    # action may have been deleted.
                    return
            if mv is not None:
                data['mvs'] = [mv]
                action.setText(MV_BTN_TEXT)
            else:
                data['mvs'] = []
                action.setText('该歌曲无 MV')
                action.setDisabled(True)
            action.setData(data)

        # artists value has not been fetched
        if data['mvs'] is None and self._fetching_mv is False:
            logger.debug('fetch song.mv for actions')
            song = data['song']
            self._fetching_mv = True
            task = run_fn(self._app.library.song_get_mv, song)
            task.add_done_callback(mv_fetched_cb)

    def _hover_artists(self, action, data):
        # pylint: disable=unnecessary-direct-lambda-call

        def artists_fetched_cb(future):
            self._fetching_artists = False
            artists = future.result()  # ignore the potential exception
            if artists:
                for artist in artists:
                    try:
                        artist_action = action.menu().addAction(artist.name)
                        # create a closure to bind variable artist
                        artist_action.triggered.connect(
                            (lambda x: lambda: self._app.browser.goto(model=x))(artist))
                    except RuntimeError:
                        # action may have been deleted.
                        return
            data['artists'] = artists or []
            action.setData(data)

        # artists value has not been fetched
        if data['artists'] is None and self._fetching_artists is False:
            logger.debug('fetch song.artists for actions')
            song = data['song']
            self._fetching_artists = True
            task = run_fn(lambda: self._app.library.song_upgrade(song).artists)
            task.add_done_callback(artists_fetched_cb)



def can_convert_chinese():
    try:
        import inlp.convert.chinese  # noqa
    except ImportError:
        return False
    return True


def convert_chinese(s: str, lang):
    """Simple to traditional, or reverse.

    Please ensure can_convert_chinese is True before invoke this function.

    Note, this feature is available only if the `inlp` package is installed.
    """

    import inlp.convert.chinese as cv  # noqa

    try:
        # I don't know what 'auto' refers to.
        assert lang in ['auto', 'cn', 'tc']

        if lang == 'cn':
            return cv.t2s(s)
        elif lang == 'tc':
            return cv.s2t(s)
        return s
    except:  # noqa
        return s



from enum import IntFlag


class ModelState(IntFlag):
    none = 0x00000000

    #: the model is created manually
    # the model identifier may not exist
    # the model fields values are not accurate
    artificial = 0x00000001

    #: the model identifier existence is proved
    # the model fields values are not accurate
    exists = 0x00000002

    #: the model identifier does not exist
    not_exists = 0x00000004

    #: the model identifier existence is proved and fields value are accurate
    # the model is a brief model
    cant_upgrade = exists | 0x00000010

    #: the model identifier existence is proved and fields value are accurate
    # the model is a normal model instead of a brief model
    upgraded = exists | 0x000000020



# -*- coding: utf-8 -*-

"""
feeluown.library
~~~~~~~~~~~~~~~~

"""
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Tuple, cast

from feeluown.media import Media, Quality
from .base import ModelType
from .models import V2SupportedModelTypes
from .flags import Flags
from .excs import MediaNotFound, ModelNotFound, NoUserLoggedIn  # noqa


class AbstractProvider(ABC):
    """
    For backward compatibility. Many providers use this as base class.
    """


class Provider:
    class meta:
        identifier: str = ''
        name: str = ''
        flags: dict = {}

    def __init__(self):
        self._user = None

    @property
    @abstractmethod
    def identifier(self):
        """provider identify"""

    @property
    @abstractmethod
    def name(self):
        """provider name"""

    def __str__(self):
        return f'provider:{self.identifier}'

    @contextmanager
    def auth_as(self, user):
        """auth as a user temporarily

        Usage::

            with auth_as(user):
                ...
        """
        old_user = self._user
        self.auth(user)
        try:
            yield
        finally:
            self.auth(old_user)

    def auth(self, user):
        """use provider as a specific user"""
        self._user = user

    def get_current_user_or_none(self):
        """
        .. versionadded: 4.0
        """
        try:
            return self.get_current_user()
        except NoUserLoggedIn:
            return None

    def search(self, *args, **kwargs):
        pass

    def use_model_v2(self, model_type: ModelType) -> bool:
        """Check whether model v2 is used for the specified model_type."""
        return Flags.model_v2 in self.meta.flags.get(model_type, Flags.none)

    def model_get(self, model_type, model_id):
        if model_type in V2SupportedModelTypes:
            if model_type == ModelType.song:
                return self.song_get(model_id)
            elif model_type == ModelType.video:
                return self.video_get(model_id)
            elif model_type == ModelType.album:
                return self.album_get(model_id)
            elif model_type == ModelType.artist:
                return self.artist_get(model_id)
            elif model_type == ModelType.playlist:
                return self.playlist_get(model_id)
        raise ModelNotFound(reason=ModelNotFound.Reason.not_supported)

    def _model_cache_get_or_fetch(self, model, cache_key):
        """Util method for getting value of cached field

        .. versionadded: 3.7.12
        """
        value, exists = model.cache_get(cache_key)
        if not exists:
            upgrade_model = self.model_get(model.meta.model_type, model.identifier)
            value, exists = upgrade_model.cache_get(cache_key)
            assert exists is True
            model.cache_set(cache_key, value)
        return value

    def song_select_media(self, song, policy=None) -> Tuple[Media, Quality.Audio]:
        """
        :raises: MediaNotFound
        """
        media, quality = self._select_media(song, policy)
        assert isinstance(quality, Quality.Audio)
        return media, quality

    def video_select_media(self, video, policy=None) -> Tuple[Media, Quality.Video]:
        """
        :raises: MediaNotFound
        """
        media, quality = self._select_media(video, policy)
        assert isinstance(quality, Quality.Video)
        return media, quality

    def _select_media(self, playable_model, policy=None):
        if ModelType(playable_model.meta.model_type) is ModelType.song:
            list_quality = self.song_list_quality
            QualityCls = Quality.Audio
            get_media = self.song_get_media
            policy = 'hq<>' if policy is None else policy
        else:
            list_quality = self.video_list_quality
            QualityCls = Quality.Video
            get_media = self.video_get_media
            policy = 'hd<>' if policy is None else policy

        # fetch available quality list
        available_q_set = set(list_quality(playable_model))
        if not available_q_set:
            raise MediaNotFound

        sorted_q_list = Quality.SortPolicy.apply(
            policy, [each.value for each in list(QualityCls)])

        # find the first available quality
        for quality in sorted_q_list:
            quality = QualityCls(quality)
            if quality not in available_q_set:
                continue
            media = get_media(playable_model, quality)
            if media is not None:
                media = cast(Media, media)
            else:
                # Media is not found for the quality. The provider show
                # a non-existing quality.
                raise MediaNotFound(
                    f'provider:{playable_model.source} has nonstandard implementation')
            return media, quality
        assert False, 'this should not happen'


ProviderV2 = Provider



import asyncio
import json
import logging

from sanic import Sanic, json as jsonify, Websocket, Request

from feeluown.app import get_app
from feeluown.serializers import serialize
from feeluown.server.pubsub import Gateway as PubsubGateway
from feeluown.server.handlers.cmd import Cmd
from feeluown.server.handlers.status import StatusHandler
from feeluown.server.handlers.player import PlayerHandler
from .jsonrpc_ import initialize, handle


logger = logging.getLogger(__name__)

# Disable sanic's logging so that it can use feeluown's logging system.
sanic_app = Sanic('FeelUOwn', configure_logging=False)


def resp(js):
    return jsonify({'code': 200, 'msg': 'ok', 'data': js})


@sanic_app.route('/api/v1/status')
async def status(request):
    cmd = Cmd('status')
    app = get_app()
    handler = StatusHandler(app)
    rv = handler.handle(cmd)
    return resp(serialize('python', rv, brief=False))


@sanic_app.post('/api/v1/player/pause')
async def pause(request):
    cmd = Cmd('pause')
    app = get_app()
    rv = PlayerHandler(app).handle(cmd)
    return resp(serialize('python', rv, brief=False))


@sanic_app.post('/api/v1/player/resume')
async def resume(request):
    cmd = Cmd('resume')
    app = get_app()
    rv = PlayerHandler(app).handle(cmd)
    return resp(serialize('python', rv, brief=False))


@sanic_app.post('/api/v1/player/play')
async def play(request):
    js = request.json
    cmd = Cmd('play', js['uri'])
    app = get_app()
    rv = PlayerHandler(app).handle(cmd)
    return resp(serialize('python', rv, brief=False))


@sanic_app.post('/rpc/v1')
async def rpcv1(request: Request):
    js = handle(request.body)
    return jsonify(js)


@sanic_app.websocket('/signal/v1')
async def signal(request, ws: Websocket):
    # TODO: 优化这个代码，比如处理连接的关闭。
    queue = asyncio.Queue()

    class Subscriber:

        def write_topic_msg(self, topic, msg):
            # TODO: 这个结构体可能会变化，需要注意一下
            queue.put_nowait(json.dumps({'topic': topic, 'data': msg, 'format': 'json'}))

    subscriber = Subscriber()

    pubsub_gateway: PubsubGateway = get_app().pubsub_gateway
    for topic in pubsub_gateway.topics:
        pubsub_gateway.link(topic, subscriber)

    while True:
        data = await queue.get()
        await ws.send(data)


async def run_web_server(host, port):
    initialize()
    server = await sanic_app.create_server(host, port, return_asyncio_server=True)
    await server.startup()
    await server.serve_forever()



# mypy: disable-error-code=type-abstract
import logging
import warnings
from functools import partial
from typing import Optional, TypeVar, List, TYPE_CHECKING

from feeluown.media import Media
from feeluown.utils.aio import run_fn, as_completed
from feeluown.utils.dispatch import Signal
from .base import SearchType, ModelType
from .provider import Provider
from .excs import MediaNotFound, ProviderAlreadyExists, ModelNotFound, ResourceNotFound
from .flags import Flags as PF
from .models import (
    ModelFlags as MF, BaseModel,
    BriefVideoModel, BriefSongModel, SongModel,
    LyricModel, VideoModel, BriefAlbumModel, BriefArtistModel
)
from .model_state import ModelState
from .provider_protocol import (
    check_flag as check_flag_impl,
    SupportsSongLyric, SupportsSongMV, SupportsSongMultiQuality,
    SupportsVideoMultiQuality, SupportsSongWebUrl, SupportsVideoWebUrl,
)

if TYPE_CHECKING:
    from .ytdl import Ytdl


logger = logging.getLogger(__name__)

FULL_SCORE = 10
MIN_SCORE = 5
T_p = TypeVar('T_p')


def raise_(e):
    raise e


def default_score_fn(origin, standby):

    # TODO: move this function to utils module
    def duration_ms_to_duration(ms):
        if not ms:  # ms is empty
            return 0
        parts = ms.split(':')
        assert len(parts) in (2, 3), f'invalid duration format: {ms}'
        if len(parts) == 3:
            h, m, s = parts
        else:
            m, s = parts
            h = 0
        return int(h) * 3600 + int(m) * 60 + int(s)

    score = FULL_SCORE
    if origin.artists_name != standby.artists_name:
        score -= 3
    if origin.title != standby.title:
        score -= 2
    if origin.album_name != standby.album_name:
        score -= 2

    if isinstance(origin, SongModel):
        origin_duration = origin.duration
    else:
        origin_duration = duration_ms_to_duration(origin.duration_ms)
    if isinstance(standby, SongModel):
        standby_duration = standby.duration
    else:
        standby_duration = duration_ms_to_duration(standby.duration_ms)
    if abs(origin_duration - standby_duration) / max(origin_duration, 1) > 0.1:
        score -= 3

    # Debug code for score function
    # print(f"{score}\t('{standby.title}', "
    #       f"'{standby.artists_name}', "
    #       f"'{standby.album_name}', "
    #       f"'{standby.duration_ms}')")
    return score


class Library:
    """Resource entrypoints."""

    def __init__(self, providers_standby=None):
        """

        :type app: feeluown.app.App
        """
        self._providers_standby = providers_standby
        self._providers = set()
        self.ytdl: Optional['Ytdl'] = None

        self.provider_added = Signal()  # emit(AbstractProvider)
        self.provider_removed = Signal()  # emit(AbstractProvider)

    def setup_ytdl(self, *args, **kwargs):
        from .ytdl import Ytdl

        self.ytdl = Ytdl(*args, **kwargs)

    def register(self, provider):
        """register provider

        :raises ProviderAlreadyExists:
        :raises ValueError:
        """
        if not isinstance(provider, Provider):
            raise ValueError('invalid provider instance')
        for _provider in self._providers:
            if _provider.identifier == provider.identifier:
                raise ProviderAlreadyExists
        self._providers.add(provider)
        self.provider_added.emit(provider)

    def deregister(self, provider) -> bool:
        """deregister provider

        .. versionchanged:: 4.0
           Do not raise exception anymore, return False instead.
        """
        if provider in self._providers:
            self._providers.remove(provider)
            self.provider_removed.emit(provider)
            return True
        return False

    def get(self, identifier) -> Optional[Provider]:
        """通过资源提供方唯一标识获取提供方实例"""
        for provider in self._providers:
            if provider.identifier == identifier:
                return provider
        return None

    def list(self) -> List[Provider]:
        """列出所有资源提供方"""
        return list(self._providers)

    def _filter(self, identifier_in=None):
        if identifier_in is None:
            return iter(self._providers)
        return filter(lambda p: p.identifier in identifier_in, self.list())

    def search(self, keyword, type_in=None, source_in=None, **kwargs):
        """search song/artist/album/playlist by keyword

        please use a_search method if you can.

        :param keyword: search keyword
        :param type_in: search type
        :param source_in: None or provider identifier list

        - TODO: support search with filters(by artist or by source)
        """
        type_in = SearchType.batch_parse(type_in) if type_in else [SearchType.so]
        for provider in self._filter(identifier_in=source_in):
            for type_ in type_in:
                try:
                    result = provider.search(keyword=keyword, type_=type_, **kwargs)
                except Exception:  # pylint: disable=broad-except
                    logger.exception('Search %s in %s failed.', keyword, provider)
                else:
                    if result is not None:
                        yield result

    async def a_search(self, keyword, source_in=None, timeout=None,
                       type_in=None,
                       **_):
        """async version of search

        TODO: add Happy Eyeballs requesting strategy if needed
        """
        type_in = SearchType.batch_parse(type_in) if type_in else [SearchType.so]

        fs = []  # future list
        for provider in self._filter(identifier_in=source_in):
            for type_ in type_in:
                future = run_fn(partial(provider.search, keyword, type_=type_))
                fs.append(future)

        for future in as_completed(fs, timeout=timeout):
            try:
                result = await future
            except:  # noqa
                logger.exception('search task failed')
                continue
            else:
                # When a provider does not implement search method, it returns None.
                if result is not None:
                    yield result

    async def a_list_song_standby_v2(self, song,
                                     audio_select_policy='>>>', source_in=None,
                                     score_fn=None, min_score=MIN_SCORE, limit=1):
        """list song standbys and their media

        .. versionadded:: 3.7.8

        """

        async def prepare_media(standby, policy):
            media = None
            try:
                media = await run_fn(self.song_prepare_media, standby, policy)
            except MediaNotFound as e:
                logger.debug(f'standby media not found: {e}')
            except:  # noqa
                logger.exception(f'get standby:{standby} media failed')
            return media

        if source_in is None:
            pvd_ids = self._providers_standby or [pvd.identifier for pvd in self.list()]
        else:
            pvd_ids = [pvd.identifier for pvd in self._filter(identifier_in=source_in)]
        if score_fn is None:
            score_fn = default_score_fn
        limit = max(limit, 1)

        q = '{} {}'.format(song.title_display, song.artists_name_display)
        standby_score_list = []  # [(standby, score), (standby, score)]
        song_media_list = []     # [(standby, media), (standby, media)]
        async for result in self.a_search(q, source_in=pvd_ids):
            if result is None:
                continue
            # Only check the first 3 songs
            for standby in result.songs:
                score = score_fn(song, standby)
                if score == FULL_SCORE:
                    media = await prepare_media(standby, audio_select_policy)
                    if media is None:
                        continue
                    logger.debug(f'find full mark standby for song:{q}')
                    song_media_list.append((standby, media))
                    if len(song_media_list) >= limit:
                        # Return as early as possible to get better performance
                        return song_media_list
                elif score >= min_score:
                    standby_score_list.append((standby, score))
        standby_pvd_id_set = {standby.source for standby, _ in standby_score_list}
        logger.debug(f"find {len(standby_score_list)} similar songs "
                     f"from {','.join(standby_pvd_id_set)}")
        # Limit try times since prapare_media is an expensive IO operation
        max_try = len(pvd_ids) * 2
        for standby, score in sorted(standby_score_list,
                                     key=lambda song_score: song_score[1],
                                     reverse=True)[:max_try]:
            media = await prepare_media(standby, audio_select_policy)
            if media is not None:
                song_media_list.append((standby, media))
                if len(song_media_list) >= limit:
                    return song_media_list
        return song_media_list

    def check_flags(self, source: str, model_type: ModelType, flags: PF) -> bool:
        """Check if a provider satisfies the specific ability for a model type

        .. note::

             Currently, we use ProviderFlags to define which ability a
             provider has. In the future, we may use typing.Protocol.
             So you should use :meth:`check_flags` method to check ability
             instead of compare provider flags directly.
        """
        provider = self.get(source)
        if provider is None:
            return False

        # Check each flag.
        for value in PF.__members__.values():
            if value in PF(flags):
                if check_flag_impl(provider, model_type, flags) is False:
                    return False
        return True

    def check_flags_by_model(self, model: BaseModel, flags: PF) -> bool:
        """Alias for check_flags"""
        warnings.warn('please use isinstance(provider, protocol_cls)')
        return self.check_flags(model.source,
                                ModelType(model.meta.model_type),
                                flags)

    # -----
    # Songs
    # -----
    def song_upgrade(self, song: BriefSongModel) -> SongModel:
        return self._model_upgrade(song)  # type: ignore

    def song_prepare_media(self, song: BriefSongModel, policy) -> Media:
        provider = self.get(song.source)
        if provider is None:
            raise MediaNotFound(f'provider({song.source}) not found')
        media = None
        if isinstance(provider, SupportsSongMultiQuality):
            try:
                media, _ = provider.song_select_media(song, policy)
            except MediaNotFound as e:
                if e.reason is MediaNotFound.Reason.check_children:
                    raise
                else:
                    media = None
        if not media:
            if self.ytdl is not None and isinstance(provider, SupportsSongWebUrl):
                song_web_url = provider.song_get_web_url(song)
                media = self.ytdl.select_audio(song_web_url, policy, source=song.source)
                found = media is not None
                logger.debug(f'ytdl select audio for {song_web_url} finished, '
                             f'found: {found}')
            if not media:
                raise MediaNotFound('provider returns empty media')
        return media

    def song_get_web_url(self, song: BriefSongModel) -> str:
        """Get song web url

        :raises ResourceNotFound: provider/song/web_url is not found
        :raises ProviderIOError: provider raises error during get_web_url

        .. versionadded:: 4.1.8
        """
        provider = self.get(song.source)
        if provider is None:
            raise ResourceNotFound(f'provider({song.source}) not found')
        if isinstance(provider, SupportsSongWebUrl):
            return provider.song_get_web_url(song)
        raise ResourceNotFound(reason=ResourceNotFound.Reason.not_supported)

    def song_prepare_mv_media(self, song: BriefSongModel, policy) -> Media:
        """

        .. versionadded:: 3.7.5
        """
        mv = self.song_get_mv(song)
        if mv is not None:
            media = self.video_prepare_media(mv, policy)
            return media
        raise MediaNotFound('provider returns empty media')

    def song_get_mv(self, song: BriefSongModel) -> Optional[VideoModel]:
        """Get the MV model of a song."""
        provider = self.get(song.source)
        if isinstance(provider, SupportsSongMV):
            return provider.song_get_mv(song)
        return None

    def song_get_lyric(self, song: BriefSongModel) -> Optional[LyricModel]:
        """Get the lyric model of a song.

        Return None when lyric does not exist instead of raising exceptions,
        because it is predictable.
        """
        provider = self.get(song.source)
        if isinstance(provider, SupportsSongLyric):
            return provider.song_get_lyric(song)
        return None

    # --------
    # Album
    # --------
    def album_upgrade(self, album: BriefAlbumModel):
        return self._model_upgrade(album)

    # --------
    # Artist
    # --------
    def artist_upgrade(self, artist: BriefArtistModel):
        return self._model_upgrade(artist)

    # --------
    # Playlist
    # --------

    def playlist_upgrade(self, playlist):
        return self._model_upgrade(playlist)

    # -------------------------
    # generic methods for model
    # -------------------------
    def model_get(self, pid, mtype, mid):
        """Get a (normal) model instance.

        :param pid: provider id
        :param mtype: model type
        :param mid: model id
        :return: model

        :raise ResourceNotFound: model does not exist
        """
        provider = self.get(pid)
        if provider is None:
            raise ModelNotFound(f'provider:{pid} not found')
        return provider.model_get(mtype, mid)

    def model_get_cover(self, model):
        """Get the cover url of model

        :param model: model which has a 'cover' field.
        :return: cover url if exists, else ''.
        """
        if MF.normal not in model.meta.flags:
            try:
                um = self._model_upgrade(model)
            except ResourceNotFound:
                return ''
        else:
            um = model
        # FIXME: remove this hack lator.
        if ModelType(model.meta.model_type) is ModelType.artist:
            cover = um.pic_url
        else:
            cover = um.cover
        return cover

    def _model_upgrade(self, model):
        """Upgrade a model to normal model.

        :raises ModelNotFound: the model does not exist

        Note you may catch ResourceNotFound exception to simplify your code.

        .. versionchanged:: 3.8.11
            Raise ModelNotFound if the model does not exist.
            Before ModelCannotUpgrade was raised.
        """
        # Return model directly if it is already a normal(upgraded) model.
        if MF.normal in model.meta.flags:
            return model

        model_type = ModelType(model.meta.model_type)
        provider = self.get(model.source)
        if provider is None:
            raise ModelNotFound(f'provider:{model.source} not found')
        try:
            upgraded_model = provider.model_get(model_type, model.identifier)
        except ModelNotFound as e:
            if e.reason is ModelNotFound.Reason.not_found:
                model.state = ModelState.not_exists
            elif e.reason is ModelNotFound.Reason.not_supported:
                model.state = ModelState.cant_upgrade
            raise
        if upgraded_model is None:  # some provider does not implement
            raise ModelNotFound(f'{provider} implementation error, it returns None :(')
        return upgraded_model

    # --------
    # Video
    # --------
    def video_upgrade(self, video):
        return self._model_upgrade(video)

    def video_prepare_media(self, video: BriefVideoModel, policy) -> Media:
        """Prepare media for video.

        :param video: either a v1 MvModel or a v2 (Brief)VideoModel.
        """
        provider = self.get(video.source)
        try:
            if isinstance(provider, SupportsVideoMultiQuality):
                media, _ = provider.video_select_media(video, policy)
                if not media:
                    raise MediaNotFound('provider returns empty media')
            else:
                raise MediaNotFound('provider or video not found')
        except MediaNotFound:
            if self.ytdl is not None and isinstance(provider, SupportsVideoWebUrl):
                video_web_url = provider.video_get_web_url(video)
                media = self.ytdl.select_video(video_web_url,
                                               policy,
                                               source=video.source)
                found = media is not None
                logger.debug(f'ytdl select video for {video_web_url} finished, '
                             f'found: {found}')
            else:
                media = None
            if not media:
                raise
        return media



from typing import TYPE_CHECKING

from PyQt5.QtCore import Qt

from feeluown.gui.widgets.labels import MessageLabel

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


async def render_error_message(app: 'GuiApp', msg: str):
    label = MessageLabel(msg, MessageLabel.ERROR)
    label.setAlignment(Qt.AlignCenter)
    app.ui.page_view.set_body(label)



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QSlider

from feeluown.app import App


class VolumeSlider(QSlider):
    def __init__(self, app: App, parent=None):
        super().__init__(parent)

        self._app = app
        self.setMinimum(0)
        self.setMaximum(100)
        self.setValue(100)
        self.sliderMoved.connect(self.on_slider_moved)
        self.setOrientation(Qt.Horizontal)

    def on_slider_moved(self, value):
        self._app.player.volume = value



"""
serializers for feeluown objects

TODO: too much code to define serializers for an object.
"""

from feeluown.app import App
from feeluown.library import AbstractProvider, SimpleSearchResult, reverse
from feeluown.player import PlaybackMode, State, Metadata
from . import PlainSerializer, PythonSerializer, \
    SerializerMeta, SimpleSerializerMixin
from .python import ListSerializer as PythonListSerializer
from .plain import ListSerializer as PlainListSerializer


class SearchSerializerMixin:
    """

    .. note::

        SearchModel isn't a standard model, it does not have identifier,
        the uri of SearchModel instance is also not so graceful, so we handle
        it as a normal object temporarily.
    """

    class Meta:
        types = (SimpleSearchResult,)

    def _get_items(self, result):
        fields = ('songs', 'albums', 'artists', 'playlists',)
        items = []
        for field in fields:
            value = getattr(result, field)
            if value:  # only append if it is not empty
                items.append((field, value))
        return items


class ProviderSerializerMixin:
    class Meta:
        types = (AbstractProvider,)

    def _get_items(self, provider):
        """
        :type provider: AbstractProvider
        """
        return [
            ('identifier', provider.identifier),
            ('uri', 'fuo://{}'.format(provider.identifier)),
            ('name', provider.name),
        ]


class AppSerializerMixin:
    class Meta:
        types = (App,)

    def _get_items(self, app):
        player = app.player
        playlist = app.playlist
        live_lyric = app.live_lyric

        repeat = playlist.playback_mode in (PlaybackMode.one_loop, PlaybackMode.loop)
        random = playlist.playback_mode == PlaybackMode.random
        items = [
            ('repeat', repeat),
            ('random', random),
            ('volume', player.volume),
            ('state', player.state.name),
        ]
        if player.state in (State.playing, State.paused) and \
                player.current_song is not None:
            items.extend([
                ('duration', player.duration),
                ('position', player.position),
                ('song', player.current_song),
                ('lyric-s', live_lyric.current_sentence),
            ])
        return items


class DictLikeSerializerMixin:
    class Meta:
        types = (Metadata,)

    def _get_items(self, metadata):
        return [(key.value, value) for key, value in metadata.items()]


# Python serializers.
#
class AppPythonSerializer(PythonSerializer,
                          AppSerializerMixin,
                          SimpleSerializerMixin,
                          metaclass=SerializerMeta):
    pass


class DictLikePythonSerializer(PythonSerializer,
                               DictLikeSerializerMixin,
                               SimpleSerializerMixin,
                               metaclass=SerializerMeta):
    pass


class ProviderPythonSerializer(PythonSerializer,
                               SimpleSerializerMixin,
                               ProviderSerializerMixin,
                               metaclass=SerializerMeta):
    pass


class SearchPythonSerializer(PythonSerializer, SearchSerializerMixin,
                             metaclass=SerializerMeta):

    def serialize(self, result):
        list_serializer = PythonListSerializer()
        json_ = {}
        for field, value in self._get_items(result):
            json_[field] = list_serializer.serialize(value)
        return json_


# Plain serializers.
#
class AppPlainSerializer(PlainSerializer,
                         AppSerializerMixin,
                         SimpleSerializerMixin,
                         metaclass=SerializerMeta):
    ...


class ProviderPlainSerializer(PlainSerializer, ProviderSerializerMixin,
                              metaclass=SerializerMeta):

    def __init__(self, **options):
        super().__init__(**options)
        self.opt_as_line = options.get('as_line', False)
        self.opt_uri_length = options.get('uri_length', '')

    def serialize(self, provider):
        """
        :type provider: AbstractProvider
        """
        items = self._get_items(provider)
        dict_ = dict(items)
        uri = dict_['uri']
        name = dict_['name']
        if self.opt_as_line or self.opt_level > 0:
            return '{uri:{uri_length}}\t# {name}'.format(
                uri=uri,
                name=name,
                uri_length=self.opt_uri_length
            )
        return self.serialize_items(items)


class SearchPlainSerializer(PlainSerializer, SearchSerializerMixin,
                            metaclass=SerializerMeta):

    def __init__(self, **options):
        super().__init__(**options)
        self.opt_uri_length = options.get('uri_length', '')

    def serialize(self, result):
        items = self._get_items(result)
        # when serialize SearchModel, we formatt it as one line when level > 1
        if self.opt_level >= 2:
            return str(result)  # I think we will never use this line format
        text_list = []
        for field, value in items:
            serializer = PlainListSerializer(
                level=self.opt_level - 1,
                fetch=False,
                uri_length=self.opt_uri_length
            )
            value_text = serializer.serialize(value)
            text_list.append(value_text)
        return '\n'.join(text_list)

    def calc_max_uri_length(self, result):
        items = self._get_items(result)
        uri_length = 0
        for field, value in items:
            for each in value:
                uri_length = max(uri_length, len(reverse(each)))
        return uri_length



from .cmd import Cmd
from .excs import HandlerException
from .base import AbstractHandler


class HelpHandler(AbstractHandler):
    cmds = 'help'

    def handle(self, cmd: Cmd):
        return self.handle_help(*cmd.args)

    def handle_help(self, cmdname: str):
        # pylint: disable=import-outside-toplevel
        from feeluown.server.dslv2 import create_dsl_parser, get_subparser  # noqa

        parser = create_dsl_parser()
        subparser = get_subparser(parser, cmdname)
        if subparser is None:
            raise HandlerException(f'help: no such cmd: {cmdname}')
        return subparser.format_help()



"""
Code are almost copied from asgiref.
https://github.com/django/asgiref/blob/6689c0a1e820a5c32532f069e945da8230d808ad/asgiref/sync.py
Lines that were changed are marked with `Note`.

Changes:

1. *2021-11-04*
   As only AsyncTosync is needed, other unrelated functions/classes are removed.
   It works well in known cases.
2. *2022-01-10*
   Since there can be only one QEventLoop instance, new_event_loop should
   create other event loop.
"""
import asyncio
import asyncio.coroutines
import functools
import inspect
import sys
import threading
import warnings
import queue
from concurrent.futures import Executor, Future, ThreadPoolExecutor
from typing import Any


if sys.version_info >= (3, 7):
    import contextvars
    get_running_loop = asyncio.get_running_loop
else:
    contextvars = None
    get_running_loop = asyncio.get_event_loop


class _WorkItem:
    """
    Represents an item needing to be run in the executor.
    Copied from ThreadPoolExecutor, but it's private, so we're not
    going to rely on importing it.
    """

    def __init__(self, future, fn, args, kwargs):
        self.future = future
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def run(self):
        if not self.future.set_running_or_notify_cancel():
            return
        try:
            result = self.fn(*self.args, **self.kwargs)
        except BaseException as exc:
            self.future.set_exception(exc)
            # Break a reference cycle with the exception 'exc'
            self = None
        else:
            self.future.set_result(result)


class CurrentThreadExecutor(Executor):
    """
    An Executor that actually runs code in the thread it is instantiated in.
    Passed to other threads running async code, so they can run sync code in
    the thread they came from.
    """

    def __init__(self):
        self._work_thread = threading.current_thread()
        self._work_queue = queue.Queue()
        self._broken = False

    def run_until_future(self, future):
        """
        Runs the code in the work queue until a result is available from the future.
        Should be run from the thread the executor is initialised in.
        """
        # Check we're in the right thread
        if threading.current_thread() != self._work_thread:
            raise RuntimeError(
                "You cannot run CurrentThreadExecutor from a different thread"
            )
        future.add_done_callback(self._work_queue.put)
        # Keep getting and running work items until we get the future we're waiting for
        # back via the future's done callback.
        try:
            while True:
                # Get a work item and run it
                work_item = self._work_queue.get()
                if work_item is future:
                    return
                work_item.run()
                del work_item
        finally:
            self._broken = True

    def submit(self, fn, *args, **kwargs):
        # Check they're not submitting from the same thread
        if threading.current_thread() == self._work_thread:
            raise RuntimeError(
                "You cannot submit onto CurrentThreadExecutor from its own thread"
            )
        # Check they're not too late or the executor errored
        if self._broken:
            raise RuntimeError("CurrentThreadExecutor already quit or is broken")
        # Add to work queue
        f = Future()
        work_item = _WorkItem(f, fn, args, kwargs)
        self._work_queue.put(work_item)
        # Return the future
        return f


def _restore_context(context):
    # Check for changes in contextvars, and set them to the current
    # context for downstream consumers
    for cvar in context:
        try:
            if cvar.get() != context.get(cvar):
                cvar.set(context.get(cvar))
        except LookupError:
            cvar.set(context.get(cvar))


def _iscoroutinefunction_or_partial(func: Any) -> bool:
    # Python < 3.8 does not correctly determine partially wrapped
    # coroutine functions are coroutine functions, hence the need for
    # this to exist. Code taken from CPython.
    if sys.version_info >= (3, 8):
        return asyncio.iscoroutinefunction(func)
    else:
        while inspect.ismethod(func):
            func = func.__func__
        while isinstance(func, functools.partial):
            func = func.func

        return asyncio.iscoroutinefunction(func)


class AsyncToSync:
    """
    Utility class which turns an awaitable that only works on the thread with
    the event loop into a synchronous callable that works in a subthread.
    If the call stack contains an async loop, the code runs there.
    Otherwise, the code runs in a new loop in a new thread.
    Either way, this thread then pauses and waits to run any thread_sensitive
    code called from further down the call stack using SyncToAsync, before
    finally exiting once the async task returns.
    """

    # Maps launched Tasks to the threads that launched them (for locals impl)
    # launch_map: "Dict[asyncio.Task[object], threading.Thread]" = {}
    # Note(cosven): we do not need launch_map since we do not use locals impl.

    # Keeps track of which CurrentThreadExecutor to use. This uses an asgiref
    # Local, not a threadlocal, so that tasks can work out what their parent used.
    #
    # Note(cosven): we can use threading.Local because we only use AsyncToSync
    # and never use SyncToAsync.
    executors = threading.local()

    def __init__(self, awaitable, force_new_loop=False):
        if not callable(awaitable) or not _iscoroutinefunction_or_partial(awaitable):
            # Python does not have very reliable detection of async functions
            # (lots of false negatives) so this is just a warning.
            warnings.warn(
                "async_to_sync was passed a non-async-marked callable", stacklevel=2
            )
        self.awaitable = awaitable
        try:
            self.__self__ = self.awaitable.__self__
        except AttributeError:
            pass
        if force_new_loop:
            # They have asked that we always run in a new sub-loop.
            self.main_event_loop = None
        else:
            try:
                self.main_event_loop = get_running_loop()
            except RuntimeError:
                # Note(cosven): we never use SyncToAsync
                #
                # There's no event loop in this thread. Look for the threadlocal if
                # we're inside SyncToAsync
                # main_event_loop_pid = getattr(
                #     SyncToAsync.threadlocal, "main_event_loop_pid", None
                # )
                # We make sure the parent loop is from the same process - if
                # they've forked, this is not going to be valid any more (#194)
                # if main_event_loop_pid and main_event_loop_pid == os.getpid():
                #     self.main_event_loop = getattr(
                #         SyncToAsync.threadlocal, "main_event_loop", None
                #     )
                # else:
                self.main_event_loop = None

    def __call__(self, *args, **kwargs):
        # You can't call AsyncToSync from a thread with a running event loop
        try:
            event_loop = get_running_loop()
        except RuntimeError:
            pass
        else:
            if event_loop.is_running():
                raise RuntimeError(
                    "You cannot use AsyncToSync in the same thread as "
                    "an async event loop - "
                    "just await the async function directly."
                )

        if contextvars is not None:
            # Wrapping context in list so it can be reassigned from within
            # `main_wrap`.
            context = [contextvars.copy_context()]
        else:
            context = None

        # Make a future for the return information
        call_result = Future()
        # Get the source thread
        source_thread = threading.current_thread()
        # Make a CurrentThreadExecutor we'll use to idle in this thread - we
        # need one for every sync frame, even if there's one above us in the
        # same thread.
        if hasattr(self.executors, "current"):
            old_current_executor = self.executors.current
        else:
            old_current_executor = None
        current_executor = CurrentThreadExecutor()
        self.executors.current = current_executor
        # Use call_soon_threadsafe to schedule a synchronous callback on the
        # main event loop's thread if it's there, otherwise make a new loop
        # in this thread.
        try:
            awaitable = self.main_wrap(
                args, kwargs, call_result, source_thread, sys.exc_info(), context
            )

            if not (self.main_event_loop and self.main_event_loop.is_running()):
                # Make our own event loop - in a new thread - and run inside that.

                # Note(cosven): There must be only one QEventLoop instance,
                # so we should use asyncio.DefaultEventLoopPolicy to create new
                # event loops.
                policy = asyncio.get_event_loop_policy()
                try:
                    asyncio.set_event_loop_policy(asyncio.DefaultEventLoopPolicy())
                    loop = asyncio.new_event_loop()
                finally:
                    asyncio.set_event_loop_policy(policy)

                loop_executor = ThreadPoolExecutor(max_workers=1)
                loop_future = loop_executor.submit(
                    self._run_event_loop, loop, awaitable
                )
                if current_executor:
                    # Run the CurrentThreadExecutor until the future is done
                    current_executor.run_until_future(loop_future)
                # Wait for future and/or allow for exception propagation
                loop_future.result()
            else:
                # Call it inside the existing loop
                self.main_event_loop.call_soon_threadsafe(
                    self.main_event_loop.create_task, awaitable
                )
                if current_executor:
                    # Run the CurrentThreadExecutor until the future is done
                    current_executor.run_until_future(call_result)
        finally:
            # Clean up any executor we were running
            if hasattr(self.executors, "current"):
                del self.executors.current
            if old_current_executor:
                self.executors.current = old_current_executor
            if contextvars is not None:
                _restore_context(context[0])

        # Wait for results from the future.
        return call_result.result()

    def _run_event_loop(self, loop, coro):
        """
        Runs the given event loop (designed to be called in a thread).
        """
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(coro)
        finally:
            try:
                # mimic asyncio.run() behavior
                # cancel unexhausted async generators
                if sys.version_info >= (3, 7, 0):
                    tasks = asyncio.all_tasks(loop)
                else:
                    tasks = asyncio.Task.all_tasks(loop)
                for task in tasks:
                    task.cancel()

                async def gather():
                    await asyncio.gather(*tasks, return_exceptions=True)

                loop.run_until_complete(gather())
                for task in tasks:
                    if task.cancelled():
                        continue
                    if task.exception() is not None:
                        loop.call_exception_handler(
                            {
                                "message": "unhandled exception during loop shutdown",
                                "exception": task.exception(),
                                "task": task,
                            }
                        )
                if hasattr(loop, "shutdown_asyncgens"):
                    loop.run_until_complete(loop.shutdown_asyncgens())
            finally:
                loop.close()
                asyncio.set_event_loop(self.main_event_loop)

    def __get__(self, parent, objtype):
        """
        Include self for methods
        """
        func = functools.partial(self.__call__, parent)
        return functools.update_wrapper(func, self.awaitable)

    async def main_wrap(
        self, args, kwargs, call_result, source_thread, exc_info, context
    ):
        """
        Wraps the awaitable with something that puts the result into the
        result/exception future.
        """
        if context is not None:
            _restore_context(context[0])

        # current_task = SyncToAsync.get_current_task()
        # self.launch_map[current_task] = source_thread
        try:
            # If we have an exception, run the function inside the except block
            # after raising it so exc_info is correctly populated.
            if exc_info[1]:
                try:
                    raise exc_info[1]
                except BaseException:
                    result = await self.awaitable(*args, **kwargs)
            else:
                result = await self.awaitable(*args, **kwargs)
        except BaseException as e:
            call_result.set_exception(e)
        else:
            call_result.set_result(result)
        finally:
            # del self.launch_map[current_task]

            if context is not None:
                context[0] = contextvars.copy_context()



from PyQt5.QtWidgets import QPushButton
from PyQt5.QtGui import QPainter, QPolygonF, QColor, QTextOption
from PyQt5.QtCore import QPointF, QRectF, Qt

from feeluown.player import PlaylistMode
from feeluown.gui.helpers import SOLARIZED_COLORS


class PlaylistButton(QPushButton):
    def __init__(self, app, parent):
        super().__init__(parent=parent)
        self._app = app

        # PlaylistButton show FM text on the button when FM mode is actiavted,
        # so update the button when mode is changed.
        self._app.playlist.mode_changed.connect(lambda *args: self.update(), weak=False)

    def paintEvent(self, e):
        width = height = 18
        painter = QPainter(self)
        painter.save()
        painter.setRenderHints(QPainter.Antialiasing)
        painter.translate((self.width() - width) // 2, (self.height() - height) // 2)

        line_height = 1.3
        line_margin = (height - 3 * line_height) / 3
        h1 = line_margin
        h2 = h1 + line_margin + line_height
        h3 = h2 + line_margin + line_height

        pen = painter.pen()
        pen.setColor(QColor("#6F6F6F"))
        pen.setWidthF(line_height)
        painter.setPen(pen)
        painter.setBrush(pen.color())

        # Draw triangle and first line
        triangle_side_length_half = line_margin * 0.6
        triangle_height = triangle_side_length_half * 1.7
        triangle = QPolygonF([QPointF(0, h1 - triangle_side_length_half),
                              QPointF(triangle_height, h1),
                              QPointF(0, h1 + triangle_side_length_half)])
        painter.drawPolygon(triangle)
        painter.drawLine(QPointF(triangle_height + triangle_side_length_half, h1),
                         QPointF(width, h1))

        # Draw second line
        painter.drawLine(QPointF(0, h2), QPointF(width, h2))

        # Draw third line, show FM text if needed
        if self._app.playlist.mode is PlaylistMode.fm:
            painter.drawLine(QPointF(0, h3), QPointF(width // 2, h3))
            painter.pen()
            pen.setColor(QColor(SOLARIZED_COLORS['blue']))
            painter.setPen(pen)
            font = painter.font()
            rect_h_half = line_margin // 2
            font.setPixelSize(int(rect_h_half * 2))
            painter.setFont(font)
            rect = QRectF(width // 2 + rect_h_half, h3 - rect_h_half,
                          width // 2 - rect_h_half, rect_h_half * 2)
            painter.drawText(rect, "FM", QTextOption(Qt.AlignCenter))
        else:
            painter.drawLine(QPointF(0, h3), QPointF(width, h3))

        painter.restore()



from .server_app import ServerApp
from .gui_app import GuiApp


class MixedApp(ServerApp, GuiApp):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self):
        super().initialize()

    def run(self):
        super().run()



import io
import sys

from PyQt5.QtCore import Qt, QTimer, pyqtSignal
from PyQt5.QtWidgets import QLineEdit, QSizePolicy

from feeluown.fuoexec import fuoexec

_KeyPrefix = 'search_'  # local storage key prefix
KeySourceIn = _KeyPrefix + 'source_in'
KeyType = _KeyPrefix + 'type'


class MagicBox(QLineEdit):
    """读取用户输入，解析执行

    ref: https://wiki.qt.io/Technical_FAQ #How can I create a one-line QTextEdit?
    """

    # this filter signal is designed for table (songs_table & albums_table)
    filter_text_changed = pyqtSignal(str)

    def __init__(self, app, parent=None):
        super().__init__(parent)

        self._app = app
        self.setPlaceholderText('搜索歌曲、歌手、专辑、用户')
        self.setToolTip('直接输入文字可以进行过滤，按 Enter 可以搜索\n'
                        '输入 >>> 前缀之后，可以执行 Python 代码\n'
                        '输入 # 前缀之后，可以过滤表格内容\n'
                        '输入 > 前缀可以执行 fuo 命令（未实现，欢迎 PR）')
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.setFixedHeight(32)
        self.setFrame(False)
        self.setAttribute(Qt.WA_MacShowFocusRect, False)
        self.setTextMargins(5, 0, 0, 0)

        self._timer = QTimer(self)
        self._cmd_text = None
        self._mode = 'cmd'  # 详见 _set_mode 函数
        self._timer.timeout.connect(self.__on_timeout)

        self.textChanged.connect(self.__on_text_edited)
        # self.textEdited.connect(self.__on_text_edited)
        self.returnPressed.connect(self.__on_return_pressed)

    def _set_mode(self, mode):
        """修改当前模式

        现在主要有两种模式：cmd 模式是正常模式；msg 模式用来展示消息通知，
        当自己处于 msg 模式下时，会 block 所有 signal
        """
        if mode == 'cmd':
            self.setReadOnly(False)
            self._timer.stop()
            self.setText(self._cmd_text or '')
            # 注意在所有操作完成之后再关闭 blockSignals
            # 然后修改当前 mode
            self.blockSignals(False)
            self._mode = mode
        elif mode == 'msg':
            self.blockSignals(True)
            if self._mode == 'cmd':
                self.setReadOnly(True)
                self._cmd_text = self.text()
                self._mode = mode

    def _exec_code(self, code):
        """执行代码并重定向代码的 stdout/stderr"""
        output = io.StringIO()
        sys.stderr = output
        sys.stdout = output
        try:
            obj = compile(code, '<string>', 'single')
            fuoexec(obj)
        except Exception as e:
            print(str(e))
        finally:
            sys.stderr = sys.__stderr__
            sys.stdout = sys.__stdout__

        text = output.getvalue()
        self._set_mode('msg')
        self.setText(text or 'No output.')
        self._timer.start(1000)
        if text:
            self._app.show_msg(text)

    def __on_text_edited(self):
        text = self.text()
        if self._mode == 'cmd':
            self._cmd_text = text
        # filter browser content if prefix starts with `#`
        # TODO: cancel unneeded filter
        if text.startswith('#'):
            self.filter_text_changed.emit(text[1:].strip())
        else:
            self.filter_text_changed.emit('')

    def __on_return_pressed(self):
        text = self.text()
        if text.startswith('>>> '):
            self._exec_code(text[4:])
        else:
            local_storage = self._app.browser.local_storage
            type_ = local_storage.get(KeyType)
            source_in = local_storage.get(KeySourceIn)
            query = {'q': text}
            if type_ is not None:
                query['type'] = type_
            if source_in is not None:
                query['source_in'] = source_in
            self._app.browser.goto(page='/search', query=query)

    def __on_timeout(self):
        self._set_mode('cmd')

    def focusInEvent(self, e):
        super().focusInEvent(e)
        self._set_mode('cmd')



import logging
from enum import IntEnum
from typing import TYPE_CHECKING, cast

from PyQt5.QtCore import Qt, QEvent
from PyQt5.QtGui import QResizeEvent
from PyQt5.QtWidgets import QWidget, QVBoxLayout

from feeluown.gui.helpers import esc_hide_widget
from feeluown.gui.widgets.frameless import ResizableFramelessContainer

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


class Mode(IntEnum):
    none = 0    # Hide all video related widgets, including controller buttons.
    fullwindow = 1  # Show video widget inside the full window coantainer.
    pip = 2     # Show video widget in a detached container (picture in picture).


class FullWindowContainer(QWidget):
    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent=parent)
        self._app = app
        self._app.installEventFilter(self)
        esc_hide_widget(self)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

    def set_body(self, widget):
        self._layout.insertWidget(0, widget)

    def showEvent(self, e):
        self.resize(self._app.size())
        super().showEvent(e)

    def eventFilter(self, obj, event):
        if self.isVisible() and obj is self._app and event.type() == QEvent.Resize:
            event = cast(QResizeEvent, event)
            self.resize(event.size())
        return False


class WatchManager:
    def __init__(self, app: 'GuiApp'):
        """video view controller

        :type app: feeluown.app.App
        :type ui: feeluown.ui.Ui
        """
        self._app = app
        self._pip_container = ResizableFramelessContainer()
        self._fullwindow_container = FullWindowContainer(app, parent=app)

        self._keep_fullwindow_mode = False
        self._keep_pip_mode = False

    def initialize(self):
        self._initialize_mpv_video_renderring()

        self._ui = self._app.ui
        self._ui.pc_panel.media_btns.toggle_video_btn.clicked.connect(
            lambda: self.keep_and_set_mode(Mode.fullwindow))
        self._app.player.media_changed.connect(self.on_media_changed, aioqueue=True)
        self._app.player.media_loaded_v2.connect(self.on_media_loaded, aioqueue=True)

        self._pip_container.setMinimumSize(200, 130)
        self._pip_container.hide()
        self._fullwindow_container.hide()

    def keep_and_set_mode(self, mode):
        mode = Mode(mode)
        if mode is Mode.fullwindow:
            self._keep_fullwindow_mode = True
        elif mode is Mode.pip:
            self._keep_pip_mode = True
        self.set_mode(mode)

    def set_mode(self, mode):
        mode = Mode(mode)  # So that user can call set_mode(0/1/2) in REPL.
        if mode is Mode.none:
            self.exit_fullwindow_mode()
            self.exit_pip_mode()
        else:
            # change current mode to mode
            #
            # if mode is same as the current mode, exit mode
            # if mode is not same as the current mode, exit
            # current mode and enter mode
            if mode is Mode.fullwindow:
                self.exit_pip_mode()
                self.enter_fullwindow_mode(go_back=self.unkeep_and_exit_fullwindow_mode)
            else:
                self.exit_fullwindow_mode()
                self.enter_pip_mode()

    def enter_fullwindow_mode(self, go_back=None):
        logger.debug("enter video-show fullwindow mode")
        video_widget = self._app.ui.mpv_widget
        if video_widget.parent() != self._fullwindow_container:
            with video_widget.change_parent():
                self._fullwindow_container.set_body(video_widget)
                video_widget.show()
                self._fullwindow_container.show()
                self._fullwindow_container.raise_()
        else:
            video_widget.show()
            self._fullwindow_container.show()
            self._fullwindow_container.raise_()

        video_widget.overlay_auto_visible = True
        video_widget.ctl_bar.clear_adhoc_btns()
        pip_btn = video_widget.ctl_bar.add_adhoc_btn('画中画')
        pip_btn.clicked.connect(lambda: self.keep_and_set_mode(Mode.pip))
        if go_back is not None:
            hide_btn = video_widget.ctl_bar.add_adhoc_btn('最小化')
            hide_btn.clicked.connect(go_back)

    def unkeep_pip_and_enter_fullwindow_mode(self):
        self._keep_pip_mode = False
        self.set_mode(Mode.fullwindow)

    def unkeep_and_exit_fullwindow_mode(self):
        self._keep_fullwindow_mode = False
        self.exit_fullwindow_mode()

    def exit_fullwindow_mode(self):
        self._fullwindow_container.hide()
        logger.debug("exit video-show fullwindow mode")

    def enter_pip_mode(self):
        """enter picture in picture mode"""
        logger.debug("enter video-show picture in picture mode")
        video_widget = self._app.ui.mpv_widget
        video_widget.overlay_auto_visible = True

        if video_widget.parent() != self._pip_container:
            with video_widget.change_parent():
                video_widget.show()
                self._pip_container.attach_widget(video_widget)
                self._pip_container.show()
        else:
            video_widget.show()
            self._pip_container.show()

        video_widget.ctl_bar.clear_adhoc_btns()
        fullscreen_btn = video_widget.ctl_bar.add_adhoc_btn('全屏')
        hide_btn = video_widget.ctl_bar.add_adhoc_btn('退出画中画')
        fullscreen_btn.clicked.connect(self.toggle_pip_fullscreen)
        hide_btn.clicked.connect(self.unkeep_pip_and_enter_fullwindow_mode)
        try:
            width = int(self._app.player._mpv.width)  # type: ignore
            height = int(self._app.player._mpv.height)  # type: ignore
        except TypeError:
            logger.exception('mpv video width/height is not a valid int')
        else:
            proper_width = max(min(width, 640), 320)
            proper_height = height * proper_width // width
            self._pip_container.resize(proper_width, proper_height)

    def exit_pip_mode(self):
        """exit picture in picture mode"""
        self._pip_container.hide()
        logger.debug("exit video-show picture in picture mode")

    def toggle_pip_fullscreen(self):
        self._pip_container.setWindowState(
            self._pip_container.windowState() ^ Qt.WindowFullScreen)

    def play_video(self, video):
        self._app.playlist.set_current_model(video)
        self.set_mode(Mode.fullwindow)

    def on_media_changed(self, media):
        if not media:
            logger.debug('media is changed to none, hide video-show')
            self.set_mode(Mode.none)

    def on_media_loaded(self, properties):
        if bool(properties['video_format']) is True:
            if self._keep_fullwindow_mode:
                self.set_mode(Mode.fullwindow)
            elif self._keep_pip_mode:
                self.set_mode(Mode.pip)
        else:
            self.set_mode(Mode.none)

    #
    # private methods
    #
    def _initialize_mpv_video_renderring(self):
        # HACK: Show mpv_widget once to initialize mpv video renderring correctly.
        # Remember to set aioqueue to true so that mpv_widget is hidden after the
        # app(Qt) event loop is started.
        self._app.ui.mpv_widget.show()
        self._app.started.connect(lambda *_: self._app.ui.mpv_widget.hide(), weak=False)



import os
from contextlib import contextmanager
from unittest.mock import MagicMock

from PyQt5.QtCore import QDir
from PyQt5.QtWidgets import QApplication, QWidget, QHBoxLayout


@contextmanager
def simple_qapp():
    qapp = QApplication([])
    yield qapp
    qapp.exec()


@contextmanager
def mock_app():
    app = MagicMock()
    yield app


def read_dark_theme_qss():
    from feeluown.gui.theme import read_resource

    pkg_root_dir = os.path.join(os.path.dirname(__file__), '..')
    icons_dir = os.path.join(pkg_root_dir, 'gui/assets/icons')
    QDir.addSearchPath('icons', icons_dir)

    qss = read_resource('common.qss')
    dark = read_resource('dark.qss')
    return qss + '\n' + dark


@contextmanager
def simple_layout(cls=QHBoxLayout, theme=''):
    with simple_qapp():
        main = QWidget()
        if theme == 'dark':
            main.setStyleSheet(read_dark_theme_qss())
        layout = cls(main)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        yield layout
        main.show()



import argparse
import textwrap
from functools import partial


def create_fmt_parser():
    fmt_parser = argparse.ArgumentParser(add_help=False)
    fmt_parser.add_argument(
        '--format',
        help="change command output format (default: plain)",
        choices=['plain', 'json'],
    )
    fmt_parser.add_argument(
        '--json',
        action='store_true',
        default=False,
        help="change command output format to json",
    )
    fmt_parser.add_argument(
        '--plain',
        action='store_true',
        default=False,
        help="change command output format to plan"
    )
    return fmt_parser


def add_common_cmds(subparsers: argparse._SubParsersAction):
    fmt_parser = create_fmt_parser()
    parents = [fmt_parser]

    add_parser = partial(subparsers.add_parser, parents=parents)

    # Create parsers.
    #
    play_parser = add_parser(
        'play',
        description=textwrap.dedent('''\
        Example:
            - fuo play fuo://netease/songs/3027393
            - fuo play "in the end"
            - fuo play 稻香-周杰伦
        '''),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    show_parser = add_parser(
        'show',
        help='显示资源详细信息',
    )
    search_parser = add_parser(
        'search',
        description=textwrap.dedent('''\
        Example:
            - fuo search hero
            - fuo search 李宗盛 source=xiami,type=artist
            - fuo search 李宗盛 [source=xiami,type=artist]
            - fuo search lizongsheng "source='xiami,qq',type=artist"
            - fuo search 李宗盛 "[source='xiami,qq',type=artist]"
        '''),
        formatter_class=argparse.RawTextHelpFormatter,
        parents=parents,
    )
    remove_parser = add_parser('remove', help='从播放列表移除歌曲')
    add_parser_ = add_parser('add', help='添加歌曲到播放列表')
    exec_parser = add_parser('exec')
    add_parser('pause', help='暂停播放')
    add_parser('resume', help='回复播放')
    add_parser('toggle', help='')
    add_parser('stop')
    add_parser('next')
    add_parser('previous')
    add_parser('list')
    add_parser('clear')
    add_parser('status')

    # Initialize parsers.
    #
    play_parser.add_argument('uri')
    show_parser.add_argument('uri')
    remove_parser.add_argument('uri')
    add_parser_.add_argument('uri', nargs='?')
    search_parser.add_argument('keyword', help='搜索关键字')
    search_parser.add_argument('-s', '--source', action='append')
    search_parser.add_argument(
        '--type',
        action='append',
        type=str,
        choices=['song', 'album', 'artist', 'video', 'playlist']
    )
    exec_parser.add_argument('code', nargs='?', help='Python 代码')


def add_cli_cmds(subparsers: argparse._SubParsersAction):
    # Generate desktop file/icon.
    subparsers.add_parser(
        'genicon',
        description='generate desktop icon'
    )


def add_server_cmds(subparsers: argparse._SubParsersAction, include_pubsub=True):
    fmt_parser = create_fmt_parser()
    subparsers.add_parser(
        'quit',
        parents=[fmt_parser],
    )
    help_parser = subparsers.add_parser(
        'help',
        parents=[fmt_parser],
    )
    help_parser.add_argument('cmdname')
    # The default value of these values should be None.
    # See set handler for more details.
    set_parser = subparsers.add_parser(
        'set',
        help='设置会话参数',
        parents=[fmt_parser],
    )
    set_parser.add_argument(
        '--rpc-version',
        type=str,
        choices=['1.0', '2.0'],
    )
    set_parser.add_argument(
        '--pubsub-version',
        type=str,
        choices=['1.0', '2.0'],
    )

    if include_pubsub is True:
        sub_parser = subparsers.add_parser(
            'sub',
            help='订阅主题消息',
            parents=[fmt_parser],
        )
        sub_parser.add_argument(
            'topics',
            nargs='?',
            help='主题名字',
        )


def _create_cli_parser() -> argparse.ArgumentParser:
    # pylint: disable=import-outside-toplevel
    from feeluown import __version__ as feeluown_version

    parser = argparse.ArgumentParser(
        description=textwrap.dedent('''\
        FeelUOwn - modern music player (daemon).

        Example:
            - fuo                        # start fuo server
            - fuo status                 # lookup server status
            - fuo play 晴天-周杰伦       # search and play
        '''),
        formatter_class=argparse.RawTextHelpFormatter,
        prog='feeluown')

    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s {}'.format(feeluown_version))
    parser.add_argument('-ns', '--no-server', action='store_true', default=False,
                        help='不运行 server')
    parser.add_argument('-nw', '--no-window', action='store_true', default=False,
                        help='不显示 GUI')

    # options about log
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='开启调试模式')
    parser.add_argument('-v', '--verbose', action='count',
                        help='输出详细的日志')
    parser.add_argument('--log-to-file', action='store_true', default=False,
                        help='将日志打到文件中')

    # XXX: 不知道能否加一个基于 regex 的 option？比如加一个
    # `--mpv-*` 的 option，否则每个 mpv 配置我都需要写一个 option？

    # TODO: 需要在文档中给出如何查看有哪些播放设备的方法
    parser.add_argument(
        '--mpv-audio-device', help='（高级选项）指定播放设备')
    return parser


def create_cli_parser() -> argparse.ArgumentParser:
    parser = _create_cli_parser()
    subparsers = parser.add_subparsers(
        dest='cmd',
    )

    add_common_cmds(subparsers)
    add_cli_cmds(subparsers)
    return parser



import sys
import os


if __name__ == '__main__':
    # Chdir first and then load feeluown, so that libmpv can be loaded correctly.
    # On macOS, the dir is changed to FeelUOwnX.app/Contents/Frameworks.
    if hasattr(sys, '_MEIPASS'):
        os.chdir(sys._MEIPASS)

    import feeluown.__main__

    os.environ.setdefault('FUO_LOG_TO_FILE', '1')
    feeluown.__main__.main()



from .metadata import MetadataFields, Metadata
from .playlist import PlaybackMode, PlaylistRepeatMode, PlaylistShuffleMode, \
    PlaylistPlayModelStage
from .base_player import State
from .mpvplayer import MpvPlayer as Player
from .playlist import PlaylistMode, Playlist
from .metadata_assembler import MetadataAssembler
from .fm import FM
from .radio import SongRadio
from .lyric import LiveLyric, parse_lyric_text, Line as LyricLine, Lyric
from .recently_played import RecentlyPlayed
from .delegate import PlayerPositionDelegate


__all__ = (
    'PlaybackMode',
    'PlaylistRepeatMode',
    'PlaylistShuffleMode',
    'PlaylistPlayModelStage',
    'State',

    'FM',
    'PlaylistMode',
    'SongRadio',

    'Player',
    'Playlist',
    'PlayerPositionDelegate',

    'Metadata',
    'MetadataFields',
    'MetadataAssembler',

    'LiveLyric',
    'parse_lyric_text',
    'LyricLine',
    'Lyric',

    'RecentlyPlayed',
)



import logging
from typing import TYPE_CHECKING

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QAbstractItemView, QMenu

from feeluown.collection import CollectionType, Collection
from feeluown.gui.widgets.textlist import TextlistModel, TextlistView

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


class CollectionListView(TextlistView):
    """
    Maybe make this a component instead of a widget.
    """
    show_collection = pyqtSignal([object])
    remove_collection = pyqtSignal([object])

    def __init__(self, app: 'GuiApp', **kwargs):
        super().__init__(**kwargs)
        self._app = app
        self.setDragDropMode(QAbstractItemView.DropOnly)
        self.setModel(CollectionListModel(self))

        self.clicked.connect(self._on_clicked)
        self._app.coll_mgr.scan_finished.connect(self.on_scan_finished)

    def on_scan_finished(self):
        model = self.model()
        assert isinstance(model, TextlistModel)
        model.clear()
        for coll in self._app.coll_mgr.listall():
            model.add(coll)

    def _on_clicked(self, index):
        collection = index.data(role=Qt.UserRole)
        self.show_collection.emit(collection)

    def contextMenuEvent(self, event):
        indexes = self.selectionModel().selectedIndexes()
        if len(indexes) != 1:
            return

        collection: Collection = self.model().data(indexes[0], Qt.UserRole)
        menu = QMenu()
        action = menu.addAction('删除此收藏集')
        action.triggered.connect(lambda: self.remove_collection.emit(collection))
        menu.exec(event.globalPos())

    # dragEnterEvent -> dragMoveEvent -> dropEvent
    def dragEnterEvent(self, e):
        """
        在这里仅判断此次 drag 的对象是否有效，如果无效，则忽略这个事件。
        之后 dropMoveEvent 也就不会接收到这个事件。
        """
        mimedata = e.mimeData()
        if mimedata.hasFormat('fuo-model/x-song') or \
           mimedata.hasFormat('fuo-model/x-album'):
            e.accept()
        else:
            e.ignore()

    def dragMoveEvent(self, e):
        """
        由于鼠标移动时，可能会移动到非 collection item 的地方，
        所以在这里判断当前位置对应的 model index 是否符合条件。
        """
        # pylint: disable=all
        index = self.indexAt(e.pos())
        if index.isValid() and index.flags() & Qt.ItemIsDropEnabled:
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        mimedata = e.mimeData()
        model = mimedata.model
        index = self.indexAt(e.pos())
        coll = index.data(Qt.UserRole)
        self._results[index.row] = (index, None)
        self.viewport().update()
        is_success = coll.add(model)
        self._results[index.row] = (index, is_success)
        self.viewport().update()
        self._result_timer.start(2000)
        e.accept()


class CollectionListModel(TextlistModel):
    def flags(self, index):
        if not index.isValid():
            return 0
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsDropEnabled
        return flags

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        item = self._items[row]
        if role == Qt.DisplayRole:
            icon = '◎  '
            if item.type in (CollectionType.sys_library, CollectionType.sys_pool):
                icon = '◉  '
            return icon + item.name
        if role == Qt.ToolTipRole:
            return item.fpath
        return super().data(index, role)



import warnings
from typing import Optional

from PyQt5.QtCore import QSize
from PyQt5.QtGui import QPainter, QImage
from PyQt5.QtWidgets import QLabel, QSizePolicy, QMenu

from feeluown.gui.drawers import PixmapDrawer
from feeluown.gui.image import open_image


class CoverLabel(QLabel):
    def __init__(self, parent=None, pixmap=None, radius=3):
        super().__init__(parent=parent)

        self._radius = radius
        self.drawer = PixmapDrawer(None, self, self._radius)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)

    def show_pixmap(self, pixmap):
        """
        .. versiondeprecated:: 3.8.11
        """
        warnings.warn('You should use show_img', DeprecationWarning)
        self.updateGeometry()
        self.update()  # Schedule a repaint to refresh the UI ASAP.

    def show_img(self, img: Optional[QImage]):
        if not img or img.isNull():
            self.drawer = PixmapDrawer(None, self, self._radius)
        else:
            self.drawer = PixmapDrawer(img, self, self._radius)
        self.updateGeometry()
        self.update()

    def paintEvent(self, e):
        """
        draw pixmap with border radius

        We found two way to draw pixmap with border radius,
        one is as follow, the other way is using bitmap mask,
        but in our practice, the mask way has poor render effects
        """
        painter = QPainter(self)
        self.drawer.draw(painter)

    def contextMenuEvent(self, e):
        if self.drawer.get_img() is None:
            return
        menu = QMenu()
        action = menu.addAction('查看原图')
        action.triggered.connect(
            lambda: open_image(self.drawer.get_img()))  # type: ignore
        menu.exec(e.globalPos())

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self.updateGeometry()

    def sizeHint(self):
        super_size = super().sizeHint()
        pixmap = self.drawer.get_pixmap()
        if pixmap is None:
            return super_size
        h = (self.width() * pixmap.height()) // pixmap.width()
        # cover label height hint can be as large as possible, since the
        # parent width has been set maximumHeigh
        w = self.width()
        return QSize(w, min(w, h))


class CoverLabelV2(CoverLabel):
    """

    .. versionadded:: 3.7.8
    """
    def __init__(self, app, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)

        self._app = app

    async def show_cover(self, url, cover_uid):
        content = await self._app.img_mgr.get(url, cover_uid)
        img = QImage()
        img.loadFromData(content)
        self.show_img(img)


if __name__ == '__main__':
    from feeluown.gui.debug import simple_layout

    with simple_layout() as layout:
        label = CoverLabel()
        layout.addWidget(label)
        label.resize(100, 100)
        label.show_img(QImage('/path/to/test.png'))



import math
from typing import Optional

from PyQt5.QtCore import Qt, QRect, QPoint, QPointF
from PyQt5.QtGui import (
    QPainter, QBrush, QPixmap, QImage, QColor, QPolygonF, QPalette,
    QPainterPath, QGuiApplication,
)
from PyQt5.QtWidgets import QWidget

from feeluown.gui.helpers import random_solarized_color, painter_save, IS_MACOS


class SizedPixmapDrawer:
    """
    Draw pixmap on a specific rect (on a fixed area).

    Note that if device_pixel_ratio is not properly set, the drawed image
    quality may be poor.
    """
    def __init__(self, img: Optional[QImage], rect: QRect, radius: int = 0):
        self._rect = rect
        self._img_old_width = rect.width()
        self._radius = radius
        self._device_pixel_ratio = QGuiApplication.instance().devicePixelRatio()

        if img is None:
            self._color = random_solarized_color()
            self._img = None
            self._pixmap = None
        else:
            self._img = img
            self._color = None
            new_img = self._scale_image(img)
            self._pixmap = QPixmap(new_img)
            self._pixmap.setDevicePixelRatio(self._device_pixel_ratio)

    def _scale_image(self, img: QImage) -> QImage:
        return img.scaledToWidth(int(self._img_old_width * self._device_pixel_ratio),
                                 Qt.SmoothTransformation)

    def get_radius(self):
        return self._radius if self._radius >= 1 else \
            self.get_rect().width() * self._radius

    def get_rect(self):
        return self._rect

    @classmethod
    def from_img_data(cls, img_data, *args, **kwargs):
        img = QImage()
        img.loadFromData(img_data)
        return cls(img, *args, **kwargs)

    def get_img(self) -> Optional[QImage]:
        return self._img

    def get_pixmap(self) -> Optional[QPixmap]:
        return self._pixmap

    def draw(self, painter: QPainter):
        with painter_save(painter):
            if self._pixmap is None:
                self._draw_random_color(painter)
            else:
                self._draw_pixmap(painter)

    def _draw_random_color(self, painter: QPainter):
        brush = QBrush(self._color)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)
        rect = self.get_rect()
        radius = self.get_radius()
        if radius == 0:
            painter.drawRect(rect)
        else:
            painter.drawRoundedRect(rect, radius, radius)

    def _draw_pixmap(self, painter: QPainter):
        assert self._pixmap is not None

        brush = QBrush(self._pixmap)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)
        radius = self.get_radius()
        size = self._pixmap.size() / self._pixmap.devicePixelRatio()
        target_rect = self.get_rect()
        y = (size.height() - target_rect.height()) // 2
        with painter_save(painter):
            painter.translate(target_rect.x(), target_rect.y())
            painter.translate(0, -y)
            rect = QRect(0, 0, target_rect.width(), size.height())
            if radius == 0:
                painter.drawRect(rect)
            else:
                painter.drawRoundedRect(rect, radius, radius)


class PixmapDrawer(SizedPixmapDrawer):
    """Draw pixmap on a widget with radius.

    The pixmap will be scaled to the width of the widget.

    TODO: rename this drawer to WidgetPixmapDrawer?
    """
    def __init__(self, img, widget: QWidget, radius: int = 0):
        """
        :param widget: a object which has width() and height() method.
        """
        super().__init__(img, widget.rect(), radius)
        self._widget = widget

    def get_rect(self):
        return self._widget.rect()

    def maybe_update_pixmap(self):
        if self._widget.width() != self._img_old_width:
            self._img_old_width = self._widget.width()
            assert self._img is not None
            new_img = self._img.scaledToWidth(self._img_old_width,
                                              Qt.SmoothTransformation)
            self._pixmap = QPixmap(new_img)

    def _draw_pixmap(self, painter: QPainter):
        self.maybe_update_pixmap()
        super()._draw_pixmap(painter)


class AvatarIconDrawer:
    def __init__(self, length, padding, fg_color=None):
        self._length = length
        self._padding = padding

        self.fg_color = fg_color

    def draw(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)

        if self.fg_color:
            painter.setPen(QColor(self.fg_color))

        diameter = self._length // 3
        # Draw circle.
        painter.drawEllipse(diameter, self._padding, diameter, diameter)
        # Draw body.
        x, y = self._padding, self._length // 2
        width, height = self._length // 2, self._length // 2
        painter.drawArc(x, y, width, height, 0, 60*16)
        painter.drawArc(x, y, width, height, 120*16, 60*16)


class PlusIconDrawer:
    def __init__(self, length, padding):
        self.top = QPoint(length//2, padding)
        self.bottom = QPoint(length//2, length - padding)
        self.left = QPoint(padding, length//2)
        self.right = QPoint(length-padding, length//2)

    def draw(self, painter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawLine(self.top, self.bottom)
        painter.drawLine(self.left, self.right)


class SearchIconDrawer:
    def __init__(self, length, padding):
        t_l_x = t_l_y = padding
        b_r_x = b_r_y = length - padding
        center = int(length * 0.6)
        self._bottom_right = QPoint(b_r_x, b_r_y)
        self._top_left = QPoint(t_l_x, t_l_y)
        self._center = QPoint(center, center)

    def draw(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawEllipse(QRect(self._top_left, self._center))
        painter.drawLine(self._center, self._bottom_right)


class TriangleIconDrawer:
    def __init__(self, length, padding, direction='up', brush=False):
        self._length = length
        self._padding = padding
        self.brush = brush

        # initialized when direction is set
        # polygon points are orgonized in anticlockwise direction
        self.triangle = QPolygonF([])
        self.set_direction(direction)

    def set_direction(self, direction):
        length = self._length
        padding = self._padding

        half = length / 2
        diameter = length - 2 * padding

        real_padding = (length - diameter) / 2
        center_top = QPointF(half, real_padding)
        center_bottom = QPointF(half, real_padding + diameter)
        left = QPointF(real_padding, half)
        right = QPointF(real_padding + diameter, half)

        d60 = diameter / 2 * 0.87  # sin60
        d30 = diameter / 2 / 2    # sin30

        if direction in ('left', 'right'):
            left_x = half - d30
            top_y = half - d60
            bottom_y = half + d60
            right_x = half + d30
        else:
            left_x = half - d60
            top_y = half - d30
            bottom_y = half + d30
            right_x = half + d60

        left_top = QPointF(left_x, top_y)
        left_bottom = QPointF(left_x, bottom_y)
        right_top = QPointF(right_x, top_y)
        right_bottom = QPointF(right_x, bottom_y)

        if direction == 'up':
            self.triangle = QPolygonF([center_top, left_bottom, right_bottom])
        elif direction == 'down':
            self.triangle = QPolygonF([center_bottom, right_top, left_top])
        elif direction == 'left':
            self.triangle = QPolygonF([left, right_bottom, right_top])
        elif direction == 'right':
            self.triangle = QPolygonF([right, left_top, left_bottom])
        else:
            raise ValueError('direction must be one of up/down/left/right')

    def draw(self, painter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        with painter_save(painter):
            if self.brush is True:
                painter.setBrush(pen.color())
            painter.drawPolygon(self.triangle)


class HomeIconDrawer:
    def __init__(self, length, padding):
        icon_length = length
        diff = 1  # root/body width diff
        h_padding = v_padding = padding

        body_left_x = h_padding + diff*2
        body_right_x = icon_length - h_padding - diff*2
        body_top_x = icon_length // 2

        self._roof = QPoint(icon_length // 2, v_padding)
        self._root_left = QPoint(h_padding, icon_length // 2 + diff)
        self._root_right = QPoint(icon_length - h_padding, icon_length // 2 + diff)

        self._body_bottom_left = QPoint(body_left_x, icon_length - v_padding)
        self._body_bottom_right = QPoint(body_right_x, icon_length - v_padding)
        self._body_top_left = QPoint(body_left_x, body_top_x)
        self._body_top_right = QPoint(body_right_x, body_top_x)

    def paint(self, painter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawLine(self._roof, self._root_left)
        painter.drawLine(self._roof, self._root_right)
        painter.drawLine(self._body_bottom_left, self._body_bottom_right)
        painter.drawLine(self._body_top_left, self._body_bottom_left)
        painter.drawLine(self._body_top_right, self._body_bottom_right)


class CalendarIconDrawer:
    def __init__(self, length, padding):
        self._body_x = self._body_y = padding
        self._body_width = length - 2 * padding
        self._radius = 3
        self._h_line_y = self._body_y + self._body_width // 4

    def paint(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        body_rect = QRect(self._body_x, self._body_x, self._body_width, self._body_width)
        painter.drawRoundedRect(body_rect, self._radius, self._radius)
        painter.drawLine(QPoint(self._body_x, self._h_line_y),
                         QPoint(self._body_x + self._body_width, self._h_line_y))


class RankIconDrawer:
    def __init__(self, length, padding):
        body = length - 2*padding
        body_2 = body // 2
        body_8 = body // 8
        body_3 = body // 3
        _top_right_x = length - padding
        _top_right_y = padding + body_8
        _bottom_left_y = padding + body - body_8

        self.p1 = QPoint(padding, _bottom_left_y)
        self.p2 = QPoint(padding + body_3, padding + body_3)
        self.p3 = QPoint(padding + body_2, padding + body_3 * 2)
        self.p4 = QPoint(_top_right_x, _top_right_y)
        self.p5 = QPoint(_top_right_x - body_3, _top_right_y)
        self.p6 = QPoint(_top_right_x, _top_right_y + body_3)

    def paint(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawLine(self.p1, self.p2)
        painter.drawLine(self.p2, self.p3)
        painter.drawLine(self.p3, self.p4)
        painter.drawLine(self.p4, self.p5)
        painter.drawLine(self.p4, self.p6)


class StarIconDrawer:
    def __init__(self, length, padding):

        radius_outer = (length - 2*padding)//2
        length_half = length // 2
        radius_inner = radius_outer // 2
        center = QPointF(length_half, length_half)
        angle = math.pi / 2

        self._star_polygon = QPolygonF()
        for _ in range(5):
            outer_point = center + QPointF(
                radius_outer * math.cos(angle),
                -radius_outer * math.sin(angle)
            )
            self._star_polygon.append(outer_point)
            inner_point = center + QPointF(
                radius_inner * math.cos(angle + math.pi/5),
                -radius_inner * math.sin(angle + math.pi/5)
            )
            self._star_polygon.append(inner_point)
            angle += 2 * math.pi / 5

    def paint(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawPolygon(self._star_polygon)


class VolumeIconDrawer:
    def __init__(self, length, padding):
        self._polygon = QPolygonF()
        self._padding = padding
        self._volume = 100

        body_length = length - 2 * padding
        body_length_4_1 = 0.25 * body_length
        body_length_2_1 = 0.5 * body_length
        body_length_4_3 = 0.75 * body_length
        body_length_3_1 = 0.33 * body_length
        body_length_3_2 = 0.66 * body_length

        left_top = QPointF(0, body_length_4_1)
        left_bottom = QPointF(0, body_length_4_3)
        mid_top = QPointF(body_length_3_1, body_length_4_1)
        mid_bottom = QPointF(body_length_3_1, body_length_4_3)
        right_top = QPointF(body_length_3_2, 0)
        right_bottom = QPointF(body_length_3_2, body_length)

        for p in (left_top, left_bottom, mid_bottom, right_bottom,
                  right_top, mid_top):
            self._polygon.append(p)

        line_right_top_y = 0.1 * body_length
        line_right_bottom_y = 0.9 * body_length
        line_left_top_y = 0.2 * body_length
        line_left_bottom_y = 0.8 * body_length
        line_left_x = 0.85 * body_length
        self._line1 = (QPointF(line_left_x, line_left_top_y),
                       QPointF(body_length, line_right_top_y))
        self._line2 = (QPointF(line_left_x, body_length_2_1),
                       QPointF(body_length, body_length_2_1))
        self._line3 = (QPointF(line_left_x, line_left_bottom_y),
                       QPointF(body_length, line_right_bottom_y))

    def get_volume(self):
        return self._volume

    def set_volume(self, volume):
        self._volume = volume

    def draw(self, painter: QPainter, palette: QPalette):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        with painter_save(painter):
            painter.translate(self._padding, self._padding)
            painter.drawPolygon(self._polygon)

            if self._volume >= 66:
                lines = (self._line1, self._line2, self._line3)
                disabled_lines = ()
            elif self._volume >= 33:
                lines = (self._line2, self._line3)
                disabled_lines = (self._line1, )
            elif self._volume > 0:
                lines = (self._line3, )
                disabled_lines = (self._line1, self._line2)
            else:
                lines = ()
                disabled_lines = (self._line1, self._line2, self._line3)
            for line in lines:
                painter.drawLine(*line)
            pen = painter.pen()
            pen.setColor(palette.color(QPalette.Disabled, QPalette.ButtonText))
            painter.setPen(pen)
            for line in disabled_lines:
                painter.drawLine(*line)


class FireIconDrawer:
    def __init__(self, length, padding):
        self.length = length
        self.padding = padding

    def paint(self, painter):
        # flake8: noqa: E501
        # The following path is copied from a SVG file.
        path = QPainterPath()
        path.moveTo(495.466219, 1024)
        path.cubicTo(256.335982, 1024, 86.993753, 856.408725, 86.993753, 625.658051)
        path.cubicTo(86.993753, 503.091298, 159.658328, 370.019367, 162.785031, 364.641439)
        path.cubicTo(169.163505, 353.385309, 181.670316, 347.006835, 194.552333, 348.507652)
        path.cubicTo(207.434349, 350.258606, 217.689934, 359.638714, 221.31691, 372.020458)
        path.cubicTo(221.567046, 372.645799, 241.202876, 447.5616, 266.966908, 488.70901)
        path.cubicTo(284.226308, 516.474132, 301.98598, 535.85969, 321.746742, 549.617182)
        path.cubicTo(308.364454, 490.83501, 298.108868, 402.411852, 314.742928, 311.862536)
        path.cubicTo(360.392654, 63.226735, 554.248234, 3.569244, 562.502729, 1.19295)
        path.cubicTo(573.133519, -1.808685, 584.639785, 0.942814, 592.769213, 8.4469)
        path.cubicTo(600.898641, 15.950986, 603.398641, 27.446525, 601.147415, 38.327452)
        path.cubicTo(600.897279, 40.078406, 568.504541, 215.423167, 636.416528, 363.87902)
        path.cubicTo(642.544866, 377.386376, 651.174556, 393.019891, 660.429596, 408.778474)
        path.cubicTo(663.056026, 386.266213, 667.43341, 364.379293, 673.561748, 342.492373)
        path.cubicTo(698.575371, 254.44442, 762.985451, 224.303004, 765.736949, 223.17739)
        path.cubicTo(776.492807, 218.424802, 788.999618, 220.675301, 798.254659, 227.929252)
        path.cubicTo(807.634768, 235.183203, 812.012152, 247.189742, 809.885993, 258.946145)
        path.cubicTo(809.510789, 261.197371, 800.505884, 321.105, 851.158472, 406.151318)
        path.cubicTo(896.071545, 482.827503, 909.078629, 532.479545, 909.078629, 628.156654)
        path.cubicTo(908.953561, 858.782259, 734.358471, 1026.248466, 495.466219, 1026.248466)
        path.closeSubpath()

        # Calculate scaling factor
        scale_factor = 1024 / (self.length - 2 * self.padding)
        painter_sf = 1 / scale_factor

        # Scale and translate the path
        transform = painter.transform()
        transform.scale(painter_sf, painter_sf)
        transform.translate(self.padding * scale_factor, self.padding * scale_factor)

        with painter_save(painter):
            painter.setTransform(transform)

            # Set the brush and pen
            pen = painter.pen()
            pen.setWidthF(1.5 * scale_factor)
            painter.setPen(pen)

            # Draw the fire shape
            painter.drawPath(path)


class EmojiIconDrawer:
    def __init__(self, emoji: str, length: int, padding: int):
        self._emoji = emoji
        self._length = length
        self._padding = padding

        self._emoji_width = self._length - 2 * self._padding

    def paint(self, painter: QPainter):
        width = self._emoji_width
        with painter_save(painter):
            painter.translate(self._padding, self._padding)
            font = painter.font()
            if IS_MACOS:
                # -4 works well on macOS when length is in range(30, 200)
                font.setPixelSize(width - 4)
            else:
                # -1 works well on KDE when length is in range(30, 200)
                font.setPixelSize(width - (self._length//20))
            painter.setFont(font)
            painter.drawText(0, 0, width, width, Qt.AlignHCenter | Qt.AlignVCenter, self._emoji)



from typing import TYPE_CHECKING

from feeluown.library import SupportsRecListDailySongs
from feeluown.gui.page_containers.table import TableContainer, Renderer
from feeluown.gui.page_containers.scroll_area import ScrollArea
from feeluown.utils.aio import run_fn
from feeluown.utils.reader import create_reader
from .template import render_error_message


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


async def render(req, **_):
    app: 'GuiApp' = req.ctx['app']
    pvd_ui = app.current_pvd_ui_mgr.get()
    if pvd_ui is None:
        return await render_error_message(app, '当前资源提供方未知，无法浏览该页面')

    provider = pvd_ui.provider

    scroll_area = ScrollArea()
    body = TableContainer(app, scroll_area)
    scroll_area.setWidget(body)
    app.ui.right_panel.set_body(scroll_area)
    if isinstance(provider, SupportsRecListDailySongs):
        songs = await run_fn(provider.rec_list_daily_songs)
        renderer = Renderer()
        await body.set_renderer(renderer)
        renderer.show_songs(create_reader(songs))
        renderer.meta_widget.show()
        renderer.meta_widget.title = '每日推荐歌曲'



import argparse
import asyncio
import logging
import os
import signal
import sys
import warnings

from feeluown.app import AppMode, create_app, create_config
from feeluown.plugin import plugins_mgr
from feeluown.utils.utils import is_port_inuse, win32_is_port_binded
from feeluown.fuoexec import fuoexec_load_rcfile, fuoexec_init
from feeluown.utils import aio
from feeluown.utils.dispatch import Signal  # noqa: E402

from .base import ensure_dirs, setup_config, setup_logger  # noqa: E402

logger = logging.getLogger(__name__)


def run_app(args: argparse.Namespace):
    args, config = before_start_app(args)
    # FIXME: qasync does not work with 'python3.11'.
    # https://github.com/CabbageDevelopment/qasync/issues/68
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        runner = asyncio.runners.Runner()
        try:
            runner.run(start_app(args, config))
        finally:
            runner.close()
    else:
        aio.run(start_app(args, config))


def before_start_app(args):
    """
    Prepare things that app depends on and initialize things which don't depend on app.
    """
    config = create_config()

    # Light scan plugins, they may define some configurations.
    plugins_mgr.light_scan()
    plugins_mgr.init_plugins_config(config)

    # Load rcfile.
    #
    # In an ideal world, users are capable to do anything in rcfile,
    # including monkeypatch, so we should load rcfile as early as possible
    fuoexec_load_rcfile(config)

    # Initialize config.
    #
    # Extract config items from args and setup config object.
    # Arg has higher priority than config. If a parameter was set both in
    # args and config, the arg can override the value.
    setup_config(args, config)

    # Precheck.
    #
    # When precheck failed, show error hint and exit.
    precheck(args, config)

    # Prepare.
    #
    # Ensure requirements, raise SystemExit if failed.
    ensure_dirs()
    setup_logger(config)
    # Ignore all warnings since it will pollute the output.
    if AppMode.cli in AppMode(config.MODE):
        warnings.filterwarnings("ignore")

    # Run.
    #
    if AppMode.gui in AppMode(config.MODE):
        if sys.platform == 'win32':
            # Enable auto scale by default so that it can work well with HiDPI display.
            os.environ.setdefault('QT_AUTO_SCREEN_SCALE_FACTOR', '1')
        elif sys.platform == 'darwin':
            # Use native event loop on macOS, so that some native service such as
            # nowplaying can work.
            os.environ.setdefault('QT_EVENT_DISPATCHER_CORE_FOUNDATION', '1')
        try:
            # HELP: QtWebEngineWidgets must be imported before a
            #   QCoreApplication instance is created.
            # TODO: add a command line option to control this import.
            import PyQt5.QtWebEngineWidgets  # type: ignore # noqa
        except ImportError:
            logger.info('import QtWebEngineWidgets failed')
        from feeluown.utils.compat import DefaultQEventLoopPolicy
        asyncio.set_event_loop_policy(DefaultQEventLoopPolicy())
    return args, config


async def start_app(args, config, sentinal=None):
    """
    The param sentinal is currently only used for unittest.
    """
    Signal.setup_aio_support()

    # create_app takes about 300ms.
    app = create_app(args, config)

    # Do fuoexec initialization before app initialization.
    fuoexec_init(app)

    # Initialize app with config.
    #
    # all objects can do initialization here. some objects may emit signal,
    # some objects may connect with others signals.
    app.initialize()

    app.initialized.emit(app)

    def sighanlder(signum, _):
        logger.info('Signal %d is received', signum)
        app.exit()

    # Handle signals.
    signal.signal(signal.SIGTERM, sighanlder)
    signal.signal(signal.SIGINT, sighanlder)

    if sentinal is None:
        sentinal: asyncio.Future = asyncio.Future()

    def shutdown(_):
        # Since about_to_shutdown signal may emit multiple times
        # (QApplication.aboutToQuit emits multiple times),
        # we should check if it is already done firstly.
        if not sentinal.done():
            sentinal.set_result(0)

    app.about_to_shutdown.connect(shutdown, weak=False)

    # GUI state must be load before running app, otherwise, it does not take effects.
    app.load_and_apply_state()
    logger.info("Load app last state...")

    # App can exit in several ways.
    #
    # GUI mode:
    # 1. QApplication.quit. QApplication.quit can be called under several circumstances
    #    1. User press CMD-Q on macOS.
    #    2. User clicks the tray icon exit button.
    # 2. SIGTERM is received.
    #
    # Daemon mode:
    # 1. Ctrl-C
    # 2. SIGTERM
    app.run()
    logger.info("App started")
    app.started.emit(app)

    await sentinal

    Signal.teardown_aio_support()


def precheck(args, config):
    # Check if there will be any errors that cause start failure.
    # If there is an error, err_msg will not be empty.
    err_msg = ''

    if AppMode.cli in AppMode(config.MODE):
        # If daemon is not started, some commands can be meaningless,
        # such as `status`, `toggle`, `next`, etc. However, some other
        # commands can still be usefule. For instance, when people
        # want to fetch a song's playable url or see the lyric of a song,
        # they may run `fuo show fuo://xxx/songs/12345`. When people
        # want to make an audition of some music, they run
        # `fuo play fuo://xxx/songs/12345`. Under these circumstances,
        # we should try to make feeluown work as they expected to.
        if args.cmd not in ('show', 'play', 'search'):
            err_msg = f"Run {args.cmd} failed, can't connect to fuo server."

    # Check if ports are in use.
    if AppMode.server in AppMode(config.MODE):
        if sys.platform == 'win32':
            host = '0.0.0.0' if config.ALLOW_LAN_CONNECT else '127.0.0.1'
            used = win32_is_port_binded(host, config.RPC_PORT) or \
                win32_is_port_binded(host, config.PUBSUB_PORT)
        else:
            # Maybe use the same checking method as windows on other platforms.
            used = is_port_inuse(config.RPC_PORT) or is_port_inuse(config.PUBSUB_PORT)
        if used:
            err_msg = (
                'App fails to start services because '
                f'either port {config.RPC_PORT} or {config.PUBSUB_PORT} '
                'was already in use. '
                'Please check if there was another FeelUOwn instance.'
            )

    if err_msg:
        if AppMode.gui in AppMode(config.MODE):
            from PyQt5.QtWidgets import QMessageBox, QApplication
            qapp = QApplication([])
            w = QMessageBox()
            w.setText(err_msg)
            # The type annotation for `finished` is wrong.
            w.finished.connect(lambda _: QApplication.quit())  # type: ignore
            w.show()
            qapp.exec()
        else:
            print(err_msg)
        sys.exit(1)



from .textbtn import TextButton  # noqa
from .login import CookiesLoginDialog  # noqa
from .selfpaint_btn import (  # noqa
    SelfPaintAbstractSquareButton, RecentlyPlayedButton,
    HomeButton, LeftArrowButton, RightArrowButton, SearchSwitchButton, SettingsButton,
    PlusButton, TriagleButton, DiscoveryButton,
    SelfPaintAbstractIconTextButton, CalendarButton, RankButton,
    StarButton, PlayPauseButton, PlayNextButton, PlayPreviousButton,
    MVButton, VolumeButton, HotButton, EmojiButton
)



from feeluown.library import (
    V2SupportedModelTypes, AlbumModel,
    SupportsAlbumSongsReader, SupportsPlaylistRemoveSong,
    SupportsArtistAlbumsReader, SupportsArtistContributedAlbumsReader,
    SupportsPlaylistSongsReader, SupportsArtistSongsReader,
)
from feeluown.utils.aio import run_fn, run_afn
from feeluown.utils.reader import create_reader
from feeluown.library import ModelType, reverse

from feeluown.gui.base_renderer import TabBarRendererMixin
from feeluown.gui.page_containers.table import Renderer
from feeluown.gui.widgets.songs import ColumnsMode
from .template import render_error_message


async def render(req, **kwargs):
    app = req.ctx['app']
    model = req.ctx['model']

    app.ui.page_view.set_body(app.ui.page_view.table_container)

    # FIXME: handle NotSupported exception here and in renderer.
    # FIXME: handle ProviderIOError and RequestException.
    if ModelType(model.meta.model_type) in V2SupportedModelTypes:
        if model.meta.model_type == ModelType.album:
            album = await run_fn(app.library.album_upgrade, model)
            tab_index = int(req.query.get('tab_index', 1))
            al_renderer = AlbumRenderer(album, tab_index)
            await app.ui.table_container.set_renderer(al_renderer)
        elif model.meta.model_type == ModelType.artist:
            artist = await run_fn(app.library.artist_upgrade, model)
            tab_index = int(req.query.get('tab_index', 1))
            ar_renderer = ArtistRenderer(artist, tab_index)
            await app.ui.table_container.set_renderer(ar_renderer)
        elif model.meta.model_type == ModelType.playlist:
            playlist = await run_fn(app.library.playlist_upgrade, model)
            pl_renderer = PlaylistRenderer(playlist)
            await app.ui.table_container.set_renderer(pl_renderer)
        else:
            assert False, "this should not be called"
            await app.ui.table_container.set_renderer(None)
    else:
        assert False, "can't render this type of model"


class ModelTabBarRendererMixin(TabBarRendererMixin):
    def render_by_tab_index(self, tab_index):
        self._app.browser.goto(model=self.model,  # type: ignore[attr-defined]
                               query={'tab_index': tab_index})


class ArtistRenderer(Renderer, ModelTabBarRendererMixin):
    def __init__(self, artist, tab_index):
        self.model = artist
        self.tab_index = tab_index
        self.tabs = [
            ('简介', ),
            ('歌曲', ),
            ('专辑', ),
            ('参与作品', ),
        ]

    async def render(self):
        artist = self.model
        tab_index = self.tab_index

        # fetch and render basic metadata
        self.meta_widget.title = await run_fn(lambda: artist.name)
        self.meta_widget.source = self._get_source_alias(artist.source)
        self.meta_widget.show()

        self.render_tab_bar()

        if tab_index == 0:
            self.show_desc(self.model.description)
        elif tab_index == 1:
            await self._show_songs()
        elif tab_index in (2, 3):
            contributed = tab_index == 3
            self.toolbar.filter_albums_needed.connect(
                lambda types: self.albums_table.model().filter_by_types(types))

            source = artist.source
            provider = self._app.library.get(source)
            if contributed is False and isinstance(provider, SupportsArtistAlbumsReader):
                reader = await run_fn(provider.artist_create_albums_rd, artist)
            elif isinstance(provider, SupportsArtistContributedAlbumsReader):
                reader = await run_fn(provider.artist_create_contributed_albums_rd, artist)  # noqa
            else:
                if contributed:
                    await render_error_message(self._app, '资源提供方不支持获取歌手贡献过的专辑')
                else:
                    await render_error_message(self._app, '资源提供方不支持获取歌手专辑')
                return
            self.toolbar.show()
            self.show_albums(reader)

        # finally, we render cover
        cover = artist.pic_url
        if cover:
            await self.show_cover(cover,
                                  reverse(artist) + '/cover',
                                  as_background=True)

    async def _show_songs(self):
        artist = self.model
        provider = self._app.library.get(artist.source)
        if not isinstance(provider, SupportsArtistSongsReader):
            await render_error_message(self._app, '资源提供方不支持获取歌手歌曲')
            return

        async def cb():
            reader = await run_fn(provider.artist_create_songs_rd, artist)
            self.__show_songs(reader)

        await cb()
        self.tabbar.show_songs_needed.connect(lambda: run_afn(cb))

    def __show_songs(self, reader):
        self.show_songs(reader=reader,
                        show_count=True,
                        columns_mode=ColumnsMode.artist)


class AlbumRenderer(Renderer, ModelTabBarRendererMixin):
    def __init__(self, album, tab_index):
        self.model = album
        self.tab_index = tab_index
        self.tabs = [
            ('简介', ),
            ('歌曲', ),
        ]

    async def render(self):
        album: AlbumModel = self.model
        tab_index = self.tab_index

        if album.released:
            self.meta_widget.released_at = album.released

        self.meta_widget.title = album.name
        self.meta_widget.creator = album.artists_name
        self.meta_widget.source = self._get_source_alias(album.source)
        self.meta_widget.show()

        self.render_tab_bar()

        if tab_index == 0:
            self.show_desc(self.model.description)
        else:
            if album.songs:
                reader = create_reader(album.songs)
            else:
                if album.song_count == 0:
                    reader = create_reader([])
                else:
                    provider = self._app.library.get(album.source)
                    if isinstance(provider, SupportsAlbumSongsReader):
                        reader = await run_fn(provider.album_create_songs_rd, album)
                    else:
                        await render_error_message(self._app, '资源提供方不支持获取专辑歌曲')
                        return
            self.meta_widget.songs_count = reader.count
            self.show_songs(reader, columns_mode=ColumnsMode.album)

        # fetch cover and description
        cover = album.cover
        if cover:
            run_afn(self.show_cover, cover, reverse(album, '/cover'))


class PlaylistRenderer(Renderer):
    def __init__(self, playlist):
        self.playlist = playlist

    async def render(self):
        playlist = self.playlist

        # show playlist title
        self.meta_widget.show()
        self.meta_widget.title = playlist.name
        self.meta_widget.source = self._get_source_alias(playlist.source)

        await self._show_songs()

        # show playlist cover
        if playlist.cover:
            run_afn(self.show_cover, playlist.cover, reverse(playlist) + '/cover')

        provider = self._app.library.get(self.playlist.source)
        if isinstance(provider, SupportsPlaylistRemoveSong):
            self.songs_table.remove_song_func = self.remove_song

    async def _show_songs(self):
        provider = self._app.library.get(self.playlist.source)
        if isinstance(provider, SupportsPlaylistSongsReader):
            reader = await run_fn(provider.playlist_create_songs_rd, self.playlist)
        else:
            await render_error_message(self._app, '资源提供方不支持获取歌单歌曲')
            return
        self.show_songs(reader=reader, show_count=True)

    def remove_song(self, song):

        async def do():
            provider = self._app.library.get(self.playlist.source)
            if await run_fn(provider.playlist_remove_song, self.playlist, song) is True:
                # Re-render songs table so that user can see that the song is removed.
                run_afn(self._show_songs)
                self._app.show_msg(f'移除歌曲 {song} 成功')
            else:
                self._app.show_msg(f'移除歌曲 {song} 失败')

        run_afn(do)



from typing import Dict, Type, TypeVar, Optional, TYPE_CHECKING

from feeluown.server.session import SessionLike

if TYPE_CHECKING:
    from feeluown.app.server_app import ServerApp


T = TypeVar('T', bound='HandlerMeta')
cmd_handler_mapping: Dict[str, 'HandlerMeta'] = {}


class HandlerMeta(type):
    def __new__(cls: Type[T], name, bases, attrs) -> T:
        klass = type.__new__(cls, name, bases, attrs)
        if 'cmds' in attrs:
            cmds = attrs['cmds']
            if isinstance(cmds, str):
                cmd_handler_mapping[cmds] = klass
            else:
                for cmd in cmds:
                    cmd_handler_mapping[cmd] = klass
        return klass


class AbstractHandler(metaclass=HandlerMeta):
    support_aio_handle = False

    def __init__(self, app: 'ServerApp', session: Optional[SessionLike] = None):
        """
        暂时不确定 session 应该设计为什么样的结构。当前主要是为了将它看作一个
        subscriber。大部分 handler 不需要使用到 session 对像，目前只有 SubHandler
        把 session 当作一个 subscriber 看待。
        """
        self._app = app
        self.session = session

        self.library = app.library
        self.player = app.player
        self.playlist = app.playlist
        self.live_lyric = app.live_lyric

    def handle(self, cmd):
        ...

    async def a_handle(self, cmd):
        ...



import locale
import logging
import time
import math
import os

from feeluown.mpv import (  # type: ignore
    MPV,
    MpvEventID,
    MpvEventEndFile,
    _mpv_set_property_string,
    _mpv_set_option_string,
    _mpv_client_api_version,
    ErrorCode,
)

from threading import Thread, RLock
from feeluown.utils.dispatch import Signal
from feeluown.media import Media, VideoAudioManifest
from .base_player import AbstractPlayer, State
from .metadata import MetadataFields, Metadata


logger = logging.getLogger(__name__)


class MpvPlayer(AbstractPlayer):
    """

    player will always play playlist current song. player will listening to
    playlist ``song_changed`` signal and change the current playback.

    todo: make me singleton
    """
    def __init__(self, _=None, audio_device=b'auto', winid=None,
                 fade=False, fade_time_ms=500, **kwargs):
        """
        :param _: keep this arg to keep backward compatibility
        """
        super().__init__(**kwargs)
        # https://github.com/cosven/FeelUOwn/issues/246
        locale.setlocale(locale.LC_NUMERIC, 'C')
        mpvkwargs = {}
        if winid is not None:
            mpvkwargs['wid'] = winid
        self._version = _mpv_client_api_version()

        # From libmpv 0.38, libmpv is not the default vo.
        # Note if vo is not set to libmpv, the video is rendered in mpv's
        # native window instead of feeluown's mpvwidget (tested on KDE+wayland).
        # On macOS, this option is optional. I believe the option is also required
        # on Windows.
        mpvkwargs['vo'] = 'libmpv'

        # set log_handler if you want to debug
        # mpvkwargs['log_handler'] = self.__log_handler
        # mpvkwargs['msg_level'] = 'all=v'
        # the default version of libmpv on Ubuntu 18.04 is (1, 25)
        self._mpv = MPV(ytdl=False,
                        input_default_bindings=True,
                        input_vo_keyboard=True,
                        **mpvkwargs)
        _mpv_set_property_string(self._mpv.handle, b'audio-device', audio_device)
        # old version libmpv(for example: (1, 20)) should set option by using
        # _mpv_set_option_string, while newer version can use _mpv_set_property_string
        _mpv_set_option_string(self._mpv.handle, b'user-agent',
                               b'Mozilla/5.0 (Windows NT 10.0; Win64; x64)')

        #: if video_format changes to None, there is no video available
        self.video_format_changed = Signal()  # Optional[str]
        self.audio_bitrate_changed = Signal()  # Optional[int], for example: 128001

        self._mpv.observe_property(
            'time-pos',
            lambda name, position: self._on_position_changed(position)
        )
        self._mpv.observe_property(
            'duration',
            lambda name, duration: self._on_duration_changed(duration)
        )
        self._mpv.observe_property(
            'video-format',
            lambda name, vformat: self._on_video_format_changed(vformat)
        )
        self._mpv.observe_property(
            'audio-bitrate',
            lambda name, bitrate: self._on_audio_bitrate_changed(bitrate)
        )
        # self._mpv.register_event_callback(lambda event: self._on_event(event))
        self._mpv._event_callbacks.append(self._on_event)
        logger.debug('Player initialize finished.')

        # this should always be `False` for fade=False,
        # because in that case `pause()` set status immediately
        self.pausing = False

        self.do_fade = fade
        if self.do_fade:
            self.fade_lock = RLock()
            self.fade_time_ms = fade_time_ms

    def shutdown(self):
        # The mpv has already been terminated.
        # The mpv can't terminate twice.
        if self._mpv.handle is not None:
            self._mpv.terminate()

    def play(self, media, video=True, metadata=None):
        if video is False:
            _mpv_set_property_string(self._mpv.handle, b'vid', b'no')
        else:
            _mpv_set_property_string(self._mpv.handle, b'vid', b'auto')

        self.media_about_to_changed.emit(self._current_media, media)
        if media is None:
            self._stop_mpv()
        else:
            logger.debug("Player will play: '%s'", media)
            if isinstance(media, Media):
                media = media
            else:  # media is a url(string)
                media = Media(media)
            self._set_http_headers(media.http_headers)
            self._set_http_proxy(media.http_proxy)
            self._stop_mpv()
            if media.manifest is None:
                url = media.url
                # Clear playlist before play next song,
                # otherwise, mpv will seek to the last position and play.
                self._mpv.play(url)
            elif isinstance(media.manifest, VideoAudioManifest):
                video_url = media.manifest.video_url
                audio_url = media.manifest.audio_url

                def add_audio():
                    try:
                        if self.current_media is media:
                            self._mpv.audio_add(audio_url)
                            self.resume()
                    finally:
                        self.media_loaded.disconnect(add_audio)

                if video is True:
                    self._mpv.play(video_url)
                    # It seems we can only add audio after the video is loaded.
                    # TODO: add method connect_once for signal
                    self.media_loaded.connect(add_audio, weak=False)
                else:
                    self._mpv.play(audio_url)
            else:
                assert False, 'Unknown manifest'
        self._current_media = media
        self.media_changed.emit(media)
        if metadata is None:
            self._current_metadata = {}
        else:
            # The metadata may be set by manual or automatic
            metadata['__setby__'] = 'manual'
            self._current_metadata = metadata
        self.metadata_changed.emit(self.current_metadata)

    def set_play_range(self, start=None, end=None):
        if self._version >= (1, 28):
            start_default, end_default = 'none', 'none'
        else:
            start_default, end_default = '0%', '100%'
        start_str = str(start) if start is not None else start_default
        end_str = str(end) if end is not None else end_default
        _mpv_set_option_string(self._mpv.handle, b'start', bytes(start_str, 'utf-8'))
        if start is not None:
            self.seeked.emit(start)
        _mpv_set_option_string(self._mpv.handle, b'end', bytes(end_str, 'utf-8'))

    def set_volume(self, max_volume: int, fade_in: bool):
        # https://bugs.python.org/issue31539#msg302699
        if os.name == "nt":
            interval = 0.02
        else:
            interval = 0.01

        freq = int(self.fade_time_ms / 1000 / interval)

        for _tick in range(freq):
            new_volume = math.ceil(
                fade_curve(_tick/freq, fade_in=fade_in)*max_volume
            )
            self.volume = new_volume
            time.sleep(interval)

    def fade_in(self):
        with self.fade_lock:
            # skip fade-in on playing
            if not self._mpv.pause:
                return

            max_volume = self.volume

            self.volume = 0
            self._resume()
            self.set_volume(max_volume, fade_in=True)

    def fade_out(self):
        with self.fade_lock:
            # skip fade-out on pause
            if self._mpv.pause or self.pausing:
                return

            max_volume = self.volume
            self.pausing = True

            self.set_volume(max_volume, fade_in=False)
            self._pause()

            self.pausing = False
            self.volume = max_volume

    def _resume(self):
        self._mpv.pause = False
        self.state = State.playing

    def _pause(self):
        self._mpv.pause = True
        self.state = State.paused

    def resume(self):
        if self.do_fade:
            Thread(target=self.fade_in).start()
        else:
            self._resume()

    def pause(self):
        if self.do_fade:
            Thread(target=self.fade_out).start()
        else:
            self._pause()

    def toggle(self):
        if self.pausing or self._mpv.pause:
            self.resume()
        else:
            self.pause()

    def stop(self):
        self._mpv.pause = True
        self.state = State.stopped
        self.play(None)
        logger.debug('Player stopped.')

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        if self._current_media:
            self._mpv.seek(position, reference='absolute')
            self._position = position
            self.seeked.emit(position)
        else:
            logger.warn("can't set position when current media is empty")

    @AbstractPlayer.volume.setter  # type: ignore
    def volume(self, value):
        super(MpvPlayer, MpvPlayer).volume.__set__(self, value)
        self._mpv.volume = self.volume

    @property
    def audio_bitrate(self):
        """
        .. versionadded: 4.1.4
        """
        return self._mpv.audio_bitrate

    @property
    def video_format(self):
        return self._video_format

    @video_format.setter
    def video_format(self, vformat):
        self._video_format = vformat
        # Note, in current implementation, the video format is changed to None
        # when switching media. So when invoking self.play(video), video_format is
        # firstly changed to None, and then changed to the real format.
        self.video_format_changed.emit(vformat)

    def _stop_mpv(self):
        # Remove current media.
        self._mpv.play("")
        # Clear the playlist so that no other media will be played.
        self._mpv.playlist_clear()

    def _on_position_changed(self, position):
        self._position = max(0, position or 0)
        self.position_changed.emit(position)

    def _on_duration_changed(self, duration):
        """listening to mpv duration change event"""
        logger.debug('Player receive duration changed signal')
        self.duration = duration

    def _on_video_format_changed(self, vformat):
        self.video_format = vformat

    def _on_audio_bitrate_changed(self, bitrate):
        self.audio_bitrate_changed.emit(bitrate)

    def _on_event(self, event):
        event_id = event['event_id']
        if event_id == MpvEventID.END_FILE:
            reason = event['event']['reason']
            logger.debug('Current song finished. reason: %d' % reason)
            if self.state != State.stopped and reason != MpvEventEndFile.ABORTED:
                self.media_finished.emit()
                if reason == MpvEventEndFile.ERROR \
                        and event['event']['error'] == ErrorCode.LOADING_FAILED:
                    self.media_loading_failed.emit()

        elif event_id == MpvEventID.FILE_LOADED:
            # If the media is a live streaming, this event may not be received.
            self.media_loaded.emit()
            self.media_loaded_v2.emit({'video_format': self._mpv.video_format})
        elif event_id == MpvEventID.METADATA_UPDATE:
            metadata = dict(self._mpv.metadata or {})  # type: ignore
            logger.debug('metadata updated to %s', metadata)
            if self._current_metadata.get('__setby__') != 'manual':
                self._current_metadata['__setby__'] = 'automatic'
                mapping = Metadata({MetadataFields.title: 'title',
                                    MetadataFields.album: 'album',
                                    MetadataFields.artists: 'artist'})
                for src, tar in mapping.items():
                    if tar in metadata:
                        value = metadata[tar]
                        if src is MetadataFields.artists:
                            value = [value]
                        self._current_metadata[src] = value
                self.metadata_changed.emit(self.current_metadata)

    def _set_http_headers(self, http_headers):
        if http_headers:
            headers = []
            for key, value in http_headers.items():
                headers.append("{}: {}".format(key, value))
            headers_text = ','.join(headers)
            headers_bytes = bytes(headers_text, 'utf-8')
            logger.info('play media with headers: %s', headers_text)
            _mpv_set_option_string(self._mpv.handle, b'http-header-fields',
                                   headers_bytes)
        else:
            _mpv_set_option_string(self._mpv.handle, b'http-header-fields',
                                   b'')

    def _set_http_proxy(self, http_proxy):
        _mpv_set_option_string(
            self._mpv.handle, b'http-proxy', bytes(http_proxy, 'utf-8'))

    def __log_handler(self, loglevel, component, message):
        print('[{}] {}: {}'.format(loglevel, component, message))


# k: factor between 0 and 1, to represent tick/fade_time
def fade_curve(k: float, fade_in: bool) -> float:
    if fade_in:
        return (1-math.cos(k*math.pi)) / 2
    else:
        return (1+math.cos(k*math.pi)) / 2



import json
from .python import PythonSerializer


class JsonSerializer(PythonSerializer):

    def __init__(self, **options):
        dump_options_keys = ['skipkeys', 'ensure_ascii', 'check_circular',
                             'allow_nan', 'cls', 'indent', 'separators',
                             'default', 'sort_keys', ]
        self.dump_options = {}
        for key in dump_options_keys:
            if key in options:
                self.dump_options[key] = options.pop(key)
        super().__init__(**options)

    def serialize(self, obj):
        dict_ = super().serialize(obj)
        return json.dumps(dict_, **self.dump_options)
